cimport numpy as np
import numpy as np
from libc.math cimport sqrt, exp, pi, floor, ceil, fabs
from cython.parallel import prange, parallel
from libc.stdlib cimport malloc, free
from libc.string cimport memset
cimport openmp
from . cimport pyfftw
from . import pyfftw
from . cimport simulation as sim

# Numpy must be initialized. When using numpy from C or Cython you must
# *ALWAYS* do that, or you will have segfaults
np.import_array()

ctypedef fused float_t:
    np.float64_t
    np.float32_t

ctypedef fused uint_t:
    np.uint64_t
    np.uint32_t

DEF FLOAT_MAX = 1.7976931348623157e+308
DEF M_1_SQRT2PI = 0.3989422804014327

ctypedef double (*loss_func)(double a) nogil

cdef double Huber_loss(double a) nogil:
    cdef double aa = fabs(a)
    if aa < 1.345:
        return 0.5 * a * a
    elif 1.345 <= aa < 3.0:
        return 1.345 * (aa - 0.6725)
    else:
        return 3.1304875

cdef double Epsilon_loss(double a) nogil:
    cdef double aa = fabs(a)
    if aa < 0.25:
        return 0.0
    elif 0.25 <= aa < 3.0:
        return aa - 0.25
    else:
        return 2.75

cdef double l2_loss(double a) nogil:
    if -3.0 < a < 3.0:
        return a * a
    else:
        return 9.0

cdef double l1_loss(double a) nogil:
    if -3.0 < a < 3.0:
        return fabs(a)
    else:
        return 3.0

cdef loss_func choose_loss(str loss):
    cdef loss_func f
    if loss == 'Epsilon':
        f = Epsilon_loss
    elif loss == 'Huber':
        f = Huber_loss
    elif loss == 'L2':
        f = l2_loss
    elif loss == 'L1':
        f = l1_loss
    else:
        raise ValueError('loss keyword is invalid')
    return f

cdef float_t min_float(float_t* array, int a) nogil:
    cdef:
        int i
        float_t mv = array[0]
    for i in range(a):
        if array[i] < mv:
            mv = array[i]
    return mv

cdef float_t max_float(float_t* array, int a) nogil:
    cdef:
        int i
        float_t mv = array[0]
    for i in range(a):
        if array[i] > mv:
            mv = array[i]
    return mv

cdef double rbf(double dsq, double h) nogil:
    return exp(-0.5 * dsq / (h * h)) * M_1_SQRT2PI

cdef void KR_frame_1d(float_t[:, ::1] I0, float_t[:, ::1] w0, uint_t[:, ::1] I_n,
                      float_t[:, ::1] W, float_t[:, :, ::1] u, float_t dj,
                      double ds_x, double h) nogil:
    cdef int X = I_n.shape[1], X0 = I0.shape[1], k, kk, k0, kk0, kk1
    cdef int dn = <int>ceil((4.0 * h) / ds_x)
    cdef double x, r

    for k in range(X):
        x = u[1, 0, k] - dj
        k0 = <int>(x / ds_x) + 1

        kk0 = k0 - dn if k0 - dn > 0 else 0
        kk1 = k0 + dn if k0 + dn < X0 else X0

        for kk in range(kk0, kk1):
            r = rbf((ds_x * kk - x) * (ds_x * kk - x), h)
            I0[0, kk] += I_n[0, k] * W[0, k] * r
            w0[0, kk] += W[0, k] * W[0, k] * r

cdef void KR_frame_2d(float_t[:, ::1] I0, float_t[:, ::1] w0, uint_t[:, ::1] I_n,
                      float_t[:, ::1] W, float_t[:, :, ::1] u, float_t di, float_t dj,
                      double ds_y, double ds_x, double h) nogil:
    cdef int Y = I_n.shape[0], X = I_n.shape[1], Y0 = I0.shape[0], X0 = I0.shape[1]
    cdef int j, k, jj, kk, j0, k0, jj0, jj1, kk0, kk1
    cdef int dn_y = <int>ceil((4.0 * h) / ds_y), dn_x = <int>ceil((4.0 * h) / ds_x)
    cdef double y, x, r

    for j in range(Y):
        for k in range(X):
            y = u[0, j, k] - di
            x = u[1, j, k] - dj
            j0 = <int>(y / ds_y) + 1
            k0 = <int>(x / ds_x) + 1

            jj0 = j0 - dn_y if j0 - dn_y > 0 else 0
            jj1 = j0 + dn_y if j0 + dn_y < Y0 else Y0
            kk0 = k0 - dn_x if k0 - dn_x > 0 else 0
            kk1 = k0 + dn_x if k0 + dn_x < X0 else X0

            for jj in range(jj0, jj1):
                for kk in range(kk0, kk1):
                    r = rbf((ds_y * jj - y) * (ds_y * jj - y) + (ds_x * kk - x) * (ds_x * kk - x), h)
                    I0[jj, kk] += I_n[j, k] * W[j, k] * r
                    w0[jj, kk] += W[j, k] * W[j, k] * r

def KR_reference(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, :, ::1] u not None,
                 float_t[::1] di not None, float_t[::1] dj not None, double ds_y, double ds_x, double h,
                 bint return_nm0=True, unsigned num_threads=1):
    r"""Generate an unabberated reference image of the sample based on the pixel
    mapping `u` and the measured data `I_n` using the Kernel regression.

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Initial sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray) : Initial sample's translations along the horizontal
            detector axis in pixels.
        ds_y (float) : Sampling interval of reference image in pixels along the
            vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along the
            horizontal axis.
        h (float) : Gaussian kernel bandwidth in pixels.
        return_nm0 (bool) : If True, also returns the lower bounds (`n0`, `m0`)
            of the reference image in pixels.
        num_threads (int) : Number of threads.

    Returns:
        Tuple[numpy.ndarray, int, int] : A tuple of three elements ('I0', 'n0',
        'm0'). The elements are the following:

        * 'I0' : Reference image array.
        * 'n0' : The lower bounds of the vertical detector axis of the reference
          image at the reference frame in pixels. Only provided if `return_nm0` is
          True.
        * 'm0' : The lower bounds of the horizontal detector axis of the reference
          image at the reference frame in pixels. Only provided if `return_nm0` is
          True.

    Notes:
        The pixel mapping `u` maps the intensity measurements from the detector
        plane to the reference plane as follows:

        .. math::
            i_{ref}[n, i, j] = u[0, i, j] + di[n], \; j_{ref}[n, i, j] = u[1, i, j] + dj[n]

        The reference image profile :math:`I_{ref}[ii, jj]` is obtained with the
        kernel regression extimator as follows:

        .. math::

            I_{ref}[i, j] = \frac{\sum_{n, i^{\prime}, j^{\prime}} K[i - i_{ref}[n,
            i^{\prime}, j^{\prime}], j - j_{ref}[n, i^{\prime}, j^{\prime}], h]
            \; W[i^{\prime}, j^{\prime}] I[n, i^{\prime}, j^{\prime}]}
            {\sum_{n, i^{\prime}, j^{\prime}} K[i - i_{ref}[n, i^{\prime}, j^{\prime}], 
            j - j_{ref}[n, i^{\prime}, j^{\prime}], h] \; W^2[i^{\prime}, j^{\prime}]}

        where :math:`K[i, j, h] = \frac{1}{\sqrt{2 \pi}} \exp(-\frac{i^2 + j^2}{h})`
        is the Gaussian kernel.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int N = I_n.shape[0], Y = I_n.shape[1], X = I_n.shape[2]
    cdef int i, j, t, k

    cdef float_t n0 = -min_float(&u[0, 0, 0], Y * X) + max_float(&di[0], N)
    cdef float_t m0 = -min_float(&u[1, 0, 0], Y * X) + max_float(&dj[0], N)
    cdef int Y0 = <int>((max_float(&u[0, 0, 0], Y * X) - min_float(&di[0], N) + n0) / ds_y) + 1
    cdef int X0 = <int>((max_float(&u[1, 0, 0], Y * X) - min_float(&dj[0], N) + m0) / ds_x) + 1
        
    cdef np.npy_intp *shape = [num_threads, Y0, X0]
    cdef float_t[:, :, ::1] I0_buf = np.PyArray_ZEROS(3, shape, type_num, 0)
    cdef float_t[:, :, ::1] W0_buf = np.PyArray_ZEROS(3, shape, type_num, 0)
    cdef np.ndarray I0 = np.PyArray_SimpleNew(2, shape + 1, type_num)

    if Y0 > 1:
        for i in prange(N, schedule='guided', num_threads=num_threads, nogil=True):
            t = openmp.omp_get_thread_num()
            KR_frame_2d(I0_buf[t], W0_buf[t], I_n[i], W, u, di[i] - n0, dj[i] - m0, ds_y, ds_x, h)
    else:
        for i in prange(N, schedule='guided', num_threads=num_threads, nogil=True):
            t = openmp.omp_get_thread_num()
            KR_frame_1d(I0_buf[t], W0_buf[t], I_n[i], W, u, dj[i] - m0, ds_x, h)

    cdef float_t[:, ::1] _I0 = I0
    cdef float_t I0_sum, W0_sum
    for k in prange(X0, schedule='guided', num_threads=num_threads, nogil=True):
        for j in range(Y0):
            I0_sum = 0.0; W0_sum = 0.0
            for i in range(<int>num_threads):
                I0_sum = I0_sum + I0_buf[i, j, k]
                W0_sum = W0_sum + W0_buf[i, j, k]

            _I0[j, k] = I0_sum / W0_sum if W0_sum > 0.0 else 1.0

    if return_nm0:
        return I0, <int>n0, <int>m0
    else:
        return I0

cdef void LOWESS_frame_1d(float_t[:, ::1] W_sum, float_t[:, :, ::1] M_mat, float_t[:, :, ::1] I0_mat,
                          float_t[:, :, ::1] W0_mat, uint_t[:, ::1] I_n, float_t[:, ::1] W,
                          float_t[:, :, ::1] u, float_t dj, double ds_x, double h) nogil:
    cdef int X = I_n.shape[1], X0 = W_sum.shape[1]
    cdef int k, kk, j0, k0, kk0, kk1
    cdef int dn = <int>ceil((4.0 * h) / ds_x)
    cdef double x, r, w

    for k in range(X):
        x = u[1, 0, k] - dj
        k0 = <int>(x / ds_x) + 1

        kk0 = k0 - dn if k0 - dn > 0 else 0
        kk1 = k0 + dn if k0 + dn < X0 else X0
    
        for kk in range(kk0, kk1):
            r = rbf((ds_x * kk - x) * (ds_x * kk - x), h)
            W_sum[0, kk] += r
            w = r / W_sum[0, kk]

            M_mat[0, kk, 1] += w * (x - M_mat[0, kk, 1])
            M_mat[0, kk, 3] += w * (x * x - M_mat[0, kk, 3])

            I0_mat[0, kk, 0] += w * (I_n[0, k] * W[0, k] - I0_mat[0, kk, 0])
            I0_mat[0, kk, 2] += w * (I_n[0, k] * W[0, k] * x - I0_mat[0, kk, 2])

            W0_mat[0, kk, 0] += w * (W[0, k] * W[0, k] - W0_mat[0, kk, 0])
            W0_mat[0, kk, 2] += w * (W[0, k] * W[0, k] * x - W0_mat[0, kk, 2])

cdef void LOWESS_frame_2d(float_t[:, ::1] W_sum, float_t[:, :, ::1] M_mat, float_t[:, :, ::1] I0_mat,
                          float_t[:, :, ::1] W0_mat, uint_t[:, ::1] I_n, float_t[:, ::1] W,
                          float_t[:, :, ::1] u, float_t di, float_t dj, double ds_y, double ds_x, double h) nogil:
    cdef int Y = I_n.shape[0], X = I_n.shape[1], Y0 = W_sum.shape[0], X0 = W_sum.shape[1]
    cdef int j, k, jj, kk, j0, k0, jj0, jj1, kk0, kk1
    cdef int dn_y = <int>ceil((4.0 * h) / ds_y), dn_x = <int>ceil((4.0 * h) / ds_x)
    cdef double y, x, r, w

    for j in range(Y):
        for k in range(X):
            y = u[0, j, k] - di
            x = u[1, j, k] - dj
            j0 = <int>(y / ds_y) + 1
            k0 = <int>(x / ds_x) + 1

            jj0 = j0 - dn_y if j0 - dn_y > 0 else 0
            jj1 = j0 + dn_y if j0 + dn_y < Y0 else Y0
            kk0 = k0 - dn_x if k0 - dn_x > 0 else 0
            kk1 = k0 + dn_x if k0 + dn_x < X0 else X0
        
            for jj in range(jj0, jj1):
                for kk in range(kk0, kk1):
                    r = rbf((ds_y * jj - y) * (ds_y * jj - y) + (ds_x * kk - x) * (ds_x * kk - x), h)
                    W_sum[jj, kk] += r
                    w = r / W_sum[jj, kk]

                    M_mat[jj, kk, 0] += w * (y - M_mat[jj, kk, 0])
                    M_mat[jj, kk, 1] += w * (x - M_mat[jj, kk, 1])
                    M_mat[jj, kk, 2] += w * (y * y - M_mat[jj, kk, 2])
                    M_mat[jj, kk, 3] += w * (x * x - M_mat[jj, kk, 3])

                    I0_mat[jj, kk, 0] += w * (I_n[j, k] * W[j, k] - I0_mat[jj, kk, 0])
                    I0_mat[jj, kk, 1] += w * (I_n[j, k] * W[j, k] * y - I0_mat[jj, kk, 1])
                    I0_mat[jj, kk, 2] += w * (I_n[j, k] * W[j, k] * x - I0_mat[jj, kk, 2])

                    W0_mat[jj, kk, 0] += w * (W[j, k] * W[j, k] - W0_mat[jj, kk, 0])
                    W0_mat[jj, kk, 1] += w * (W[j, k] * W[j, k] * y - W0_mat[jj, kk, 1])
                    W0_mat[jj, kk, 2] += w * (W[j, k] * W[j, k] * x - W0_mat[jj, kk, 2])

def LOWESS_reference(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, :, ::1] u not None,
                     float_t[::1] di not None, float_t[::1] dj not None, double ds_y, double ds_x, double h,
                     bint return_nm0=True, unsigned num_threads=1):
    r"""Generate an unabberated reference image of the sample based on the
    pixel mapping `u` and the measured data `I_n` using the Local Weighted
    Linear Regression (LOWESS).

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Initial sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray) : Initial sample's translations along the horizontal
            detector axis in pixels.
        ds_y (float) : Sampling interval of reference image in pixels along the
            vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along the
            horizontal axis.
        h (float) : Gaussian kernel bandwidth in pixels.
        return_nm0 (bool) : If True, also returns the lower bounds (`n0`, `m0`)
            of the reference image in pixels.
        num_threads (int) : Number of threads.

    Returns:
        Tuple[numpy.ndarray, int, int] : A tuple of three elements ('I0', 'n0',
        'm0'). The elements are the following:

        * 'I0' : Reference image array.
        * 'n0' : The lower bounds of the vertical detector axis of the reference
          image at the reference frame in pixels. Only provided if `return_nm0` is
          True.
        * 'm0' : The lower bounds of the horizontal detector axis of the reference
          image at the reference frame in pixels. Only provided if `return_nm0` is
          True.

    Notes:
        The pixel mapping `u` maps the intensity measurements from the
        detector plane to the reference plane as follows:

        .. math::
            i_{ref}[n, i, j] = u[0, i, j] + di[n], \; j_{ref}[n, i, j] = u[1, i, j] + dj[n]

        The reference image profile :math:`I_{ref}[ii, jj]` is obtained
        with the LOWESS regression extimator as follows:

        .. math::

            I_{ref}[i, j] = \frac{\sum_{n, i^{\prime}, j^{\prime}} K[i - i_{ref}[n,
            i^{\prime}, j^{\prime}], j - j_{ref}[n, i^{\prime}, j^{\prime}], h]
            \; r_{IW}[n, a^{IW}, b^{IW}_i, b^{IW}_j]}
            {\sum_{n, i^{\prime}, j^{\prime}} K[i - i_{ref}[n, i^{\prime}, j^{\prime}], 
            j - j_{ref}[n, i^{\prime}, j^{\prime}], h] \;
            r_{WW}[n, a^{WW}, b^{WW}_i, b^{WW}_j]}

        where :math:`r_{IW}[n, a, b_i, b_j]` and :math:`r_{WW}[n, a, b_i, b_j]`
        are the residuals at the pixel :math:`i, j` with linear coefficients
        defined with the least squares approach:

        .. math::

            r_{IW}[n, a, b_i, b_j] = I[n, i^{\prime}, j^{\prime}] W[i^{\prime}, j^{\prime}]
            - a - b_i (i - i_{ref}[n, i^{\prime}, j^{\prime}) - b_j
            (j - j_{ref}[n, i^{\prime}, j^{\prime})

        .. math::

            r_{WW}[n, a, b_i, b_j] = W^2[i^{\prime}, j^{\prime}]
            - a - b_i (i - i_{ref}[n, i^{\prime}, j^{\prime}) - b_j
            (j - j_{ref}[n, i^{\prime}, j^{\prime})

        and :math:`K[i, j, h] = \frac{1}{\sqrt{2 \pi}} \exp(-\frac{i^2 + j^2}{h})`
        is the Gaussian kernel.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int N = I_n.shape[0], Y = I_n.shape[1], X = I_n.shape[2]
    cdef int i, j, t, k

    cdef float_t n0 = -min_float(&u[0, 0, 0], Y * X) + max_float(&di[0], N)
    cdef float_t m0 = -min_float(&u[1, 0, 0], Y * X) + max_float(&dj[0], N)
    cdef int Y0 = <int>((max_float(&u[0, 0, 0], Y * X) - min_float(&di[0], N) + n0) / ds_y) + 1
    cdef int X0 = <int>((max_float(&u[1, 0, 0], Y * X) - min_float(&dj[0], N) + m0) / ds_x) + 1
        
    cdef np.npy_intp *shape = [num_threads, Y0, X0, 4]
    cdef float_t[:, :, ::1] W_buf = np.PyArray_ZEROS(3, shape, type_num, 0)
    cdef float_t[:, :, :, ::1] M_buf = np.PyArray_ZEROS(4, shape, type_num, 0)
    shape[3] = 3
    cdef float_t[:, :, :, ::1] I0_buf = np.PyArray_ZEROS(4, shape, type_num, 0)
    cdef float_t[:, :, :, ::1] W0_buf = np.PyArray_ZEROS(4, shape, type_num, 0)
    cdef np.ndarray I0 = np.PyArray_SimpleNew(2, shape + 1, type_num)

    if Y0 > 1:
        for i in prange(N, schedule='guided', num_threads=num_threads, nogil=True):
            t = openmp.omp_get_thread_num()
            LOWESS_frame_2d(W_buf[t], M_buf[t], I0_buf[t], W0_buf[t], I_n[i], W, u,
                            di[i] - n0, dj[i] - m0, ds_y, ds_x, h)
    else:
        for i in prange(N, schedule='guided', num_threads=num_threads, nogil=True):
            t = openmp.omp_get_thread_num()
            LOWESS_frame_1d(W_buf[t], M_buf[t], I0_buf[t], W0_buf[t], I_n[i], W, u,
                            dj[i] - m0, ds_x, h)

    cdef float_t[:, ::1] _I0 = I0
    cdef float_t W_sum, w, betta_y, betta_x, var_y, var_x, I0_pred, W0_pred
    cdef float_t *M_mat
    cdef float_t *I0_mat
    cdef float_t *W0_mat

    with nogil, parallel(num_threads=num_threads):
        M_mat = <float_t *>malloc(4 * sizeof(float_t))
        I0_mat = <float_t *>malloc(3 * sizeof(float_t))
        W0_mat = <float_t *>malloc(3 * sizeof(float_t))

        for k in prange(X0, schedule='guided'):

            for j in range(Y0):
                W_sum = 0.0
                memset(M_mat, 0, 4 * sizeof(float_t))
                memset(I0_mat, 0, 3 * sizeof(float_t))
                memset(W0_mat, 0, 3 * sizeof(float_t))

                for i in range(<int>num_threads):
                    if W_buf[i, j, k] > 0.0:
                        W_sum = W_sum + W_buf[i, j, k]
                        w = W_buf[i, j, k] / W_sum

                        for t in range(4):
                            M_mat[t] = M_mat[t] + w * (M_buf[i, j, k, t] - M_mat[t])
                        
                        for t in range(3):
                            I0_mat[t] = I0_mat[t] + w * (I0_buf[i, j, k, t] - I0_mat[t])
                            W0_mat[t] = W0_mat[t] + w * (W0_buf[i, j, k, t] - W0_mat[t])

                var_y = M_mat[2] - M_mat[0] * M_mat[0]
                var_x = M_mat[3] - M_mat[1] * M_mat[1]

                betta_y = (I0_mat[1] - I0_mat[0] * M_mat[0]) / var_y if var_y > 0.0 else 0.0
                betta_x = (I0_mat[2] - I0_mat[0] * M_mat[1]) / var_x if var_x > 0.0 else 0.0

                I0_pred = I0_mat[0] - betta_y * (ds_y * j - M_mat[0]) - betta_x * (ds_x * k - M_mat[1])

                betta_y = (W0_mat[1] - W0_mat[0] * M_mat[0]) / var_y if var_y > 0.0 else 0.0
                betta_x = (W0_mat[2] - W0_mat[0] * M_mat[1]) / var_x if var_x > 0.0 else 0.0

                W0_pred = W0_mat[0] - betta_y * (ds_y * j - M_mat[0]) - betta_x * (ds_x * k - M_mat[1])

                _I0[j, k] = I0_pred / W0_pred if W0_pred > 0.0 else 1.0

        free(M_mat); free(I0_mat); free(W0_mat)

    if return_nm0:
        return I0, <int>n0, <int>m0
    else:
        return I0

cdef double FVU_interp(uint_t[:, :, ::1] I_n, float_t W, float_t[:, ::1] I0, float_t[::1] di, float_t[::1] dj, int j, int k,
                       float_t ux, float_t uy, double ds_y, double ds_x, double sigma, loss_func f) nogil:
    """Return fraction of variance unexplained between the validation set I and trained
    profile I0. Find the predicted values at the points (y, x) with bilinear interpolation.
    """
    cdef int N = I_n.shape[0], Y0 = I0.shape[0], X0 = I0.shape[1]
    cdef int i, y0, y1, x0, x1
    cdef double y, x, dy, dx, I0_bi, err = 0.0

    for i in range(N):
        y = (ux - di[i]) / ds_y
        x = (uy - dj[i]) / ds_x

        if y <= 0.0:
            dy = 0.0; y0 = 0; y1 = 0
        elif y >= Y0 - 1.0:
            dy = 0.0; y0 = Y0 - 1; y1 = Y0 - 1
        else:
            dy = y - floor(y)
            y0 = <int>floor(y); y1 = y0 + 1

        if x <= 0.0:
            dx = 0.0; x0 = 0; x1 = 0
        elif x >= X0 - 1.0:
            dx = 0.0; x0 = X0 - 1; x1 = X0 - 1
        else:
            dx = x - floor(x)
            x0 = <int>floor(x); x1 = x0 + 1

        I0_bi = (1.0 - dy) * (1.0 - dx) * I0[y0, x0] + \
                (1.0 - dy) * dx * I0[y0, x1] + \
                dy * (1.0 - dx) * I0[y1, x0] + \
                dy * dx * I0[y1, x1]
        err += f((<double>I_n[i, j, k] - W * I0_bi) / sigma)
    
    return err / N

cdef double FVU_interp_tr(float_t[:, ::1] errors, uint_t[:, ::1] I_n, float_t[:, ::1] W,
                          float_t[:, ::1] I0, float_t[:, :, ::1] u, float_t di0, float_t dj0, float_t di,
                          float_t dj, double ds_y, double ds_x, double sigma, loss_func f) nogil:
    """Return fraction of variance unexplained between the validation set I and trained
    profile I0. Find the predicted values at the points (y, x) with bilinear interpolation.
    """
    cdef int Y = I_n.shape[0], X = I_n.shape[1], Y0 = I0.shape[0], X0 = I0.shape[1]
    cdef int j, k, y0, y1, x0, x1
    cdef double y, x, dy, dx, I0_bi, err0, err1, err = 0.0

    for j in range(Y):
        for k in range(X):
            y = (u[0, j, k] - di0) / ds_y
            x = (u[1, j, k] - dj0) / ds_x

            if y <= 0.0:
                dy = 0.0; y0 = 0; y1 = 0
            elif y >= Y0 - 1.0:
                dy = 0.0; y0 = Y0 - 1; y1 = Y0 - 1
            else:
                dy = y - floor(y)
                y0 = <int>floor(y); y1 = y0 + 1

            if x <= 0.0:
                dx = 0.0; x0 = 0; x1 = 0
            elif x >= X0 - 1.0:
                dx = 0.0; x0 = X0 - 1; x1 = X0 - 1
            else:
                dx = x - floor(x)
                x0 = <int>floor(x); x1 = x0 + 1

            I0_bi = (1.0 - dy) * (1.0 - dx) * I0[y0, x0] + \
                    (1.0 - dy) * dx * I0[y0, x1] + \
                    dy * (1.0 - dx) * I0[y1, x0] + \
                    dy * dx * I0[y1, x1]
            err0 = f((<double>I_n[j, k] - W[j, k] * I0_bi) / sigma)

            y = (u[0, j, k] - di0) / ds_y
            x = (u[1, j, k] - dj0) / ds_x

            if y <= 0.0:
                dy = 0.0; y0 = 0; y1 = 0
            elif y >= Y0 - 1.0:
                dy = 0.0; y0 = Y0 - 1; y1 = Y0 - 1
            else:
                dy = y - floor(y)
                y0 = <int>floor(y); y1 = y0 + 1

            if x <= 0.0:
                dx = 0.0; x0 = 0; x1 = 0
            elif x >= X0 - 1.0:
                dx = 0.0; x0 = X0 - 1; x1 = X0 - 1
            else:
                dx = x - floor(x)
                x0 = <int>floor(x); x1 = x0 + 1

            I0_bi = (1.0 - dy) * (1.0 - dx) * I0[y0, x0] + \
                    (1.0 - dy) * dx * I0[y0, x1] + \
                    dy * (1.0 - dx) * I0[y1, x0] + \
                    dy * dx * I0[y1, x1]
            err1 = f((<double>I_n[j, k] - W[j, k] * I0_bi) / sigma)

            err += (errors[j, k] - err0 + err1)

    return err / (Y * X)

cdef void pm_gsearcher(uint_t[:, :, ::1] I_n, float_t[:, ::1] W, float_t[:, ::1] I0, float_t[:, :, ::1] u,
                       float_t[:, ::1] derrs, float_t[::1] di, float_t[::1] dj, int j, int k, double sw_y, double sw_x,
                       unsigned wsize, double ds_y, double ds_x, double sigma, loss_func f) nogil:
    cdef double err, err0, uy_min = 0.0, ux_min = 0.0, err_min=FLOAT_MAX, ux, uy 
    cdef double dsw_y = 2.0 * sw_y / (wsize - 1), dsw_x = 2.0 * sw_x / (wsize - 1)
    cdef int ii, jj

    err0 = FVU_interp(I_n, W[j, k], I0, di, dj, j, k, u[0, j, k],
                      u[1, j, k], ds_y, ds_x, sigma, f)

    for ii in range(<int>wsize if dsw_y > 0.0 else 1):
        uy = dsw_y * (ii - 0.5 * (wsize - 1))
        for jj in range(<int>wsize if dsw_x > 0.0 else 1):
            ux = dsw_x * (jj - 0.5 * (wsize - 1))
            err = FVU_interp(I_n, W[j, k], I0, di, dj, j, k, u[0, j, k] + uy,
                             u[1, j, k] + ux, ds_y, ds_x, sigma, f)

            if err < err_min:
                uy_min = uy; ux_min = ux; err_min = err

    u[0, j, k] += uy_min; u[1, j, k] += ux_min
    derrs[j, k] = err0 - err_min if err_min < err0 else 0.0

def pm_gsearch(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
               float_t[:, :, ::1] u0 not None, float_t[::1] di not None, float_t[::1] dj not None,
               double sw_y, double sw_x, unsigned grid_size, double ds_y, double ds_x, double sigma,
               str loss='Huber', unsigned num_threads=1):
    r"""Update the pixel mapping by minimizing mean-squared-error
    (MSE). Perform a grid search within the search window of `sw_y`,
    `sw_x` size along the vertical and fast axes accordingly in order to
    minimize the MSE at each point of the detector grid separately.

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Initial sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray) : Initial sample's translations along the horizontal
            detector axis in pixels.
        sw_y (float) : Search window size in pixels along the vertical detector
            axis.
        sw_x (float) : Search window size in pixels along the horizontal detector
            axis.
        grid_size (int) :  Grid size along one of the detector axes. The grid
            shape is then (grid_size, grid_size).
        ds_y (float) : Sampling interval of reference image in pixels along the
            vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along the
            horizontal axis.
        sigma (float) : The standard deviation of `I_n`.
        loss (str) : Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.

        num_threads (int) : Number of threads.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray] : A tuple of two elements ('u', 'derr').
        The elements are the following:

        * 'u' : Updated pixel mapping array.
        * 'derr' : Error decrease for each pixel in the detector grid.

    Notes:
        The error metric as a function of pixel mapping displacements
        is given by:

        .. math::

            \varepsilon_{pm}[i, j, i^{\prime}, j^{\prime}] = \frac{1}{N}
            \sum_{n = 0}^N f\left( \frac{I[n, i, j] - W[i, j]
            I_{ref}[u[0, i, j] + i^{\prime} - di[n],
            u[1, i, j] + j^{\prime} - dj[n]]}{\sigma} \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int Y = I_n.shape[1], X = I_n.shape[2], j, k

    cdef np.npy_intp *u_shape = [2, Y, X]
    cdef np.ndarray u = np.PyArray_SimpleNew(3, u_shape, type_num)
    cdef np.ndarray derr = np.PyArray_ZEROS(2, u_shape + 1, type_num, 0)
    cdef float_t[:, :, ::1] _u = u
    cdef float_t[:, ::1] _derr = derr

    for k in prange(X, schedule='guided', num_threads=num_threads, nogil=True):
        for j in range(Y):
            _u[0, j, k] = u0[0, j, k]; _u[1, j, k] = u0[1, j, k]
            if W[j, k] > 0.0:
                pm_gsearcher(I_n, W, I0, _u, _derr, di, dj, j, k, sw_y, sw_x,
                             grid_size, ds_y, ds_x, sigma, f)

    return u, derr

cdef void pm_rsearcher(uint_t[:, :, ::1] I_n, float_t[:, ::1] W, float_t[:, ::1] I0, gsl_rng *r, float_t[:, :, ::1] u,
                       float_t[:, ::1] derrs, float_t[::1] di, float_t[::1] dj, int j, int k, double sw_y, double sw_x,
                       unsigned N, double ds_y, double ds_x, double sigma, loss_func f) nogil:
    cdef double err, err0, err_min=FLOAT_MAX, uy_min = 0.0, ux_min = 0.0, ux, uy
    cdef int ii

    err0 = FVU_interp(I_n, W[j, k], I0, di, dj, j, k, u[0, j, k],
                      u[1, j, k], ds_y, ds_x, sigma, f)

    for ii in range(<int>N):
        uy = 2.0 * sw_y * (gsl_rng_uniform(r) - 0.5)
        ux = 2.0 * sw_x * (gsl_rng_uniform(r) - 0.5)

        err = FVU_interp(I_n, W[j, k], I0, di, dj, j, k, u[0, j, k] + uy,
                         u[1, j, k] + ux, ds_y, ds_x, sigma, f)
        if err < err_min:
            uy_min = uy; ux_min = ux; err_min = err

    u[0, j, k] += uy_min; u[1, j, k] += ux_min
    derrs[j, k] = err0 - err_min if err_min < err0 else 0.0

def pm_rsearch(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
               float_t[:, :, ::1] u0 not None, float_t[::1] di not None, float_t[::1] dj not None,
               double sw_y, double sw_x, unsigned n_trials, unsigned long seed, double ds_y, double ds_x, double sigma,
               str loss='Huber', unsigned num_threads=1):
    r"""Update the pixel mapping by minimizing mean-squared-error (MSE).
    Perform a random search within the search window of `sw_y`, `sw_x` size
    along the vertical and fast axes accordingly in order to minimize the MSE
    at each point of the detector grid separately.

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Initial sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray) : Initial sample's translations along the horizontal
            detector axis in pixels.
        sw_y (float) : Search window size in pixels along the vertical detector
            axis.
        sw_x (float) : Search window size in pixels along the horizontal detector
            axis.
        n_trials (int) : Number of points generated at each pixel of the detector grid.
        seed (int) : Specify seed for the random number generation.
        ds_y (float) : Sampling interval of reference image in pixels along the
            vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along the
            horizontal axis.
        sigma (float) : The standard deviation of `I_n`.
        loss (str) : Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.

        num_threads (int) : Number of threads.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray] : A tuple of two elements ('u', 'derr').
        The elements are the following:

        * 'u' : Updated pixel mapping array.
        * 'derr' : Error decrease for each pixel in the detector grid.

    Notes:
        The error metric as a function of pixel mapping displacements
        is given by:

        .. math::

            \varepsilon_{pm}[i, j, i^{\prime}, j^{\prime}] = \frac{1}{N}
            \sum_{n = 0}^N f\left( \frac{I[n, i, j] - W[i, j]
            I_{ref}[u[0, i, j] + i^{\prime} - di[n],
            u[1, i, j] + j^{\prime} - dj[n]]}{\sigma} \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int Y = I_n.shape[1], X = I_n.shape[2], j, k

    cdef np.npy_intp *u_shape = [2, Y, X]
    cdef np.ndarray u = np.PyArray_SimpleNew(3, u_shape, type_num)
    cdef np.ndarray derr = np.PyArray_ZEROS(2, u_shape + 1, type_num, 0)
    cdef float_t[:, :, ::1] _u = u
    cdef float_t[:, ::1] _derr = derr

    cdef gsl_rng *r_master = gsl_rng_alloc(gsl_rng_mt19937)
    gsl_rng_set(r_master, seed)
    cdef unsigned long thread_seed
    cdef gsl_rng *r

    with nogil, parallel(num_threads=num_threads):
        r = gsl_rng_alloc(gsl_rng_mt19937)
        thread_seed = gsl_rng_get(r_master)
        gsl_rng_set(r, thread_seed)

        for k in prange(X, schedule='guided'):
            for j in range(Y):
                _u[0, j, k] = u0[0, j, k]; _u[1, j, k] = u0[1, j, k]
                if W[j, k] > 0.0:
                    pm_rsearcher(I_n, W, I0, r, _u, _derr, di, dj, j, k, sw_y, sw_x,
                                 n_trials, ds_y, ds_x, sigma, f)

        gsl_rng_free(r)

    gsl_rng_free(r_master)

    return u, derr

cdef void pm_devolver(uint_t[:, :, ::1] I_n, float_t[:, ::1] W, float_t[:, ::1] I0, gsl_rng *r, float_t[:, :, ::1] u,
                      float_t[:, ::1] derrs, float_t[::1] di, float_t[::1] dj, int j, int k, double sw_y, double sw_x,
                      unsigned NP, unsigned n_iter, double CR, double F, double ds_y, double ds_x, double sigma, loss_func f) nogil:
    cdef double err0, err, err_min = FLOAT_MAX
    cdef int ii, jj, n, a, b
    cdef double u_min[2]
    cdef double sw[2]
    cdef double *pop = <double *>malloc(2 * NP * sizeof(double))
    cdef double *cost = <double *>malloc(NP * sizeof(double))
    cdef double *new_pop = <double *>malloc(2 * NP * sizeof(double))

    sw[0] = sw_y; sw[1] = sw_x
    err0 = FVU_interp(I_n, W[j, k], I0, di, dj, j, k, u[0, j, k],
                      u[1, j, k], ds_y, ds_x, sigma, f)

    for ii in range(<int>NP):
        pop[2 * ii] = 2.0 * sw_y * (gsl_rng_uniform(r) - 0.5)
        pop[2 * ii + 1] = 2.0 * sw_x * (gsl_rng_uniform(r) - 0.5)

        cost[ii] = FVU_interp(I_n, W[j, k], I0, di, dj, j, k, u[0, j, k] + pop[2 * ii],
                             u[1, j, k] + pop[2 * ii + 1], ds_y, ds_x, sigma, f)
        
        if cost[ii] < err_min:
            u_min[0] = pop[2 * ii]; u_min[1] = pop[2 * ii + 1]; err_min = cost[ii]

    for n in range(<int>n_iter):
        for ii in range(<int>NP):
            a = gsl_rng_uniform_int(r, NP)
            while a == ii:
                a = gsl_rng_uniform_int(r, NP)
            
            b = gsl_rng_uniform_int(r, NP)
            while b == ii or b == a:
                b = gsl_rng_uniform_int(r, NP)

            jj = gsl_rng_uniform_int(r, 2)
            if gsl_rng_uniform(r) < CR:
                new_pop[2 * ii + jj] = u_min[jj] + F * (pop[2 * a + jj] - pop[2 * b + jj])
                if new_pop[2 * ii + jj] > sw[jj]: new_pop[2 * ii + jj] = sw[jj]
                if new_pop[2 * ii + jj] < -sw[jj]: new_pop[2 * ii + jj] = -sw[jj]
            else:
                new_pop[2 * ii + jj] = pop[2 * ii + jj]
            jj = (jj + 1) % 2
            new_pop[2 * ii + jj] = u_min[jj] + F * (pop[2 * a + jj] - pop[2 * b + jj])
            if new_pop[2 * ii + jj] > sw[jj]: new_pop[2 * ii + jj] = sw[jj]
            if new_pop[2 * ii + jj] < -sw[jj]: new_pop[2 * ii + jj] = -sw[jj]

            err = FVU_interp(I_n, W[j, k], I0, di, dj, j, k, u[0, j, k] + new_pop[2 * ii],
                             u[1, j, k] + new_pop[2 * ii + 1], ds_y, ds_x, sigma, f)

            if err < cost[ii]:
                cost[ii] = err
                if err < err_min:
                    u_min[0] = new_pop[2 * ii]; u_min[1] = new_pop[2 * ii + 1]; err_min = err
            else:
                new_pop[2 * ii] = pop[2 * ii]; new_pop[2 * ii + 1] = pop[2 * ii + 1]
            
        for ii in range(2 * <int>NP):
            pop[ii] = new_pop[ii]

    free(pop); free(new_pop); free(cost)

    u[0, j, k] += u_min[0]; u[1, j, k] += u_min[1]
    derrs[j, k] = err0 - err_min if err_min < err0 else 0.0

def pm_devolution(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
                  float_t[:, :, ::1] u0 not None, float_t[::1] di not None, float_t[::1] dj not None,
                  double sw_y, double sw_x, unsigned pop_size, unsigned n_iter, unsigned long seed,
                  double ds_y, double ds_x, double sigma, double F=0.75, double CR=0.7, str loss='Huber',
                  unsigned num_threads=1):
    r"""Update the pixel mapping by minimizing mean-squared-error (MSE). Perform
    a differential evolution within the search window of `sw_y`, `sw_x` size along
    the vertical and fast axes accordingly in order to minimize the MSE at each
    point of the detector grid separately.

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Initial sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray) : Initial sample's translations along the horizontal
            detector axis in pixels.
        sw_y (float) : Search window size in pixels along the vertical detector
            axis.
        sw_x (float) : Search window size in pixels along the horizontal detector
            axis.
        pop_size (int) : The total population size. Must be greater or equal to 4.
        n_iter (int) : The maximum number of generations over which the entire
            population is evolved.
        seed (int) : Specify seed for the random number generation.
        ds_y (float) : Sampling interval of reference image in pixels along the
            vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along the
            horizontal axis.
        sigma (float) : The standard deviation of `I_n`.
        F (float) : The mutation constant. In the literature this is also known as
            differential weight. If specified as a float it should be in the
            range [0, 2].
        CR (float) : The recombination constant, should be in the range [0, 1]. In
            the literature this is also known as the crossover probability.
        loss (str) : Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.

        num_threads (int) : Number of threads.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray] : A tuple of two elements ('u', 'derr').
        The elements are the following:

        * 'u' : Updated pixel mapping array.
        * 'derr' : Error decrease for each pixel in the detector grid.

    Notes:
        The error metric as a function of pixel mapping displacements
        is given by:

        .. math::

            \varepsilon_{pm}[i, j, i^{\prime}, j^{\prime}] = \frac{1}{N}
            \sum_{n = 0}^N f\left( \frac{I[n, i, j] - W[i, j]
            I_{ref}[u[0, i, j] + i^{\prime} - di[n],
            u[1, i, j] + j^{\prime} - dj[n]]}{\sigma} \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')
    if pop_size < 4:
        raise ValueError('Population size must be greater or equal to 4.')
    if F < 0.0 or F > 2.0:
        raise ValueError('The mutation constant F must be in the interval [0.0, 2.0].')
    if CR < 0.0 or CR > 1.0:
        raise ValueError('The recombination constant CR must be in the interval [0.0, 1.0].')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int Y = I_n.shape[1], X = I_n.shape[2], j, k

    cdef np.npy_intp *u_shape = [2, Y, X]
    cdef np.ndarray u = np.PyArray_SimpleNew(3, u_shape, type_num)
    cdef np.ndarray derr = np.PyArray_ZEROS(2, u_shape + 1, type_num, 0)
    cdef float_t[:, :, ::1] _u = u
    cdef float_t[:, ::1] _derr = derr

    cdef gsl_rng *r_master = gsl_rng_alloc(gsl_rng_mt19937)
    gsl_rng_set(r_master, seed)
    cdef unsigned long thread_seed
    cdef gsl_rng *r

    with nogil, parallel(num_threads=num_threads):
        r = gsl_rng_alloc(gsl_rng_mt19937)
        thread_seed = gsl_rng_get(r_master)
        gsl_rng_set(r, thread_seed)

        for k in prange(X, schedule='guided'):
            for j in range(Y):
                _u[0, j, k] = u0[0, j, k]; _u[1, j, k] = u0[1, j, k]
                if W[j, k] > 0.0:
                    pm_devolver(I_n, W, I0, r, _u, _derr, di, dj, j, k, sw_y, sw_x,
                                pop_size, n_iter, CR, F, ds_y, ds_x, sigma, f)

        gsl_rng_free(r)

    gsl_rng_free(r_master)

    return u, derr

cdef void tr_updater(float_t[:, ::1] errors, uint_t[:, ::1] I_n, float_t[:, ::1] W, float_t[:, ::1] I0,
                     float_t[:, :, ::1] u, float_t *di, float_t *dj, double sw_y, double sw_x,
                     unsigned wsize, double ds_y, double ds_x, double sigma, loss_func f) nogil:
    cdef double di_min = 0.0, dj_min = 0.0, err_min=FLOAT_MAX, dii, djj
    cdef double dsw_y = 2.0 * sw_y / (wsize - 1), dsw_x = 2.0 * sw_x / (wsize - 1)
    cdef int ii, jj

    for ii in range(<int>wsize if dsw_y > 0.0 else 1):
        dii = dsw_y * (ii - 0.5 * (wsize - 1))
        for jj in range(<int>wsize if dsw_x > 0.0 else 1):
            djj = dsw_x * (jj - 0.5 * (wsize - 1))
            err = FVU_interp_tr(errors, I_n, W, I0, u, di[0], dj[0],
                                di[0] + dii, dj[0] + djj, ds_y, ds_x, sigma, f)

            if err < err_min:
                di_min = dii; dj_min = djj; err_min = err

    di[0] += di_min; dj[0] += dj_min

def tr_gsearch(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
               float_t[:, :, ::1] u not None, float_t[::1] di not None, float_t[::1] dj not None,
               double sw_y, double sw_x, unsigned grid_size, double ds_y, double ds_x, double sigma,
               str loss='Huber', unsigned num_threads=1) -> np.ndarray:
    r"""Update the sample pixel translations by minimizing total mean-squared-error
    (:math:$MSE_{total}$). Perform a grid search within the search window of
    `sw_y` size in pixels for sample translations along the vertical axis, and
    of `sw_x` size in pixels for sample translations along the horizontal axis in
    order to minimize the total MSE.

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Initial sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray) : Initial sample's translations along the horizontal
            detector axis in pixels.
        sw_y (float) : Search window size in pixels along the vertical detector
            axis.
        sw_x (float) : Search window size in pixels along the horizontal detector
            axis.
        grid_size (int) : Grid size along one of the detector axes. The grid shape
            is then (grid_size, grid_size).
        ds_y (float) : Sampling interval of reference image in pixels along the
            vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along the
            horizontal axis.
        sigma (float) : The standard deviation of `I_n`.
        loss (str) : Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.

        num_threads (int) : Number of threads.

    Returns:
        numpy.ndarray : Updated sample pixel translations.

    Notes:
        The error metric as a function of sample shifts is given by:

        .. math::

            \varepsilon_{tr}[n, di^{\prime}, dj^{\prime}] = \frac{1}{Y X}
            \sum_{i = 0}^Y \sum_{j = 0}^Y f\left( \frac{I[n, i, j] - W[i, j]
            I_{ref}[u[0, i, j] - di[n] - di^{\prime},
            u[1, i, j] - dj[n] - dj^{\prime}]}{\sigma} \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int N = I_n.shape[0], Y = I_n.shape[1], X = I_n.shape[2], i, j, k

    cdef np.npy_intp *buf_shape = [Y, X]
    cdef float_t[:, ::1] errors = np.PyArray_SimpleNew(2, buf_shape, type_num)

    cdef np.npy_intp *dij_shape = [N, 2]
    cdef np.ndarray dij = np.PyArray_SimpleNew(2, dij_shape, type_num)
    cdef float_t[:, ::1] _dij = dij

    for k in prange(X, schedule='guided', num_threads=num_threads, nogil=True):
        for j in range(Y):
            if W[j, k] > 0.0:
                errors[j, k] = FVU_interp(I_n, W[j, k], I0, di, dj, j, k,
                                          u[0, j, k], u[1, j, k], ds_y, ds_x, sigma, f)
            else:
                errors[j, k] = 0.0

    for i in prange(N, schedule='guided', num_threads=num_threads, nogil=True):
        _dij[i, 0] = di[i]; _dij[i, 1] = dj[i]
        tr_updater(errors, I_n[i], W, I0, u, &_dij[i, 0], &_dij[i, 1],
                   sw_y, sw_x, grid_size, ds_y, ds_x, sigma, f)

    return dij

def pm_errors(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
              float_t[:, :, ::1] u not None, float_t[::1] di not None, float_t[::1] dj not None,
              double ds_y, double ds_x, double sigma, str loss='Huber', unsigned num_threads=1) -> np.ndarray:
    r"""Return the residuals for the pixel mapping fit.

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Sample's translations along the vertical detector
            axis in pixels.
        dj (numpy.ndarray) : Sample's translations along the fast detector
            axis in pixels.
        ds_y (float) : Sampling interval of reference image in pixels along
            the vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along
            the horizontal axis.
        sigma (float) : The standard deviation of `I_n`.
        loss (str): Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.
        num_threads (int): Number of threads.

    Returns:
        numpy.ndarray : Residual profile of the pixel mapping fit.

    See Also:
        :func:`pyrost.bin.pm_gsearch` : Description of error metric which
        is being minimized.

    Notes:
        The error metric is given by:

        .. math::

            \varepsilon_{pm}[i, j] = \frac{1}{N}
            \sum_{n = 0}^N f\left( \frac{I[n, i, j] - W[i, j]
            I_{ref}[u[0, i, j] - di[n], u[1, i, j] - dj[n]]}{\sigma}
            \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int Y = I_n.shape[1], X = I_n.shape[2], j, k

    cdef np.ndarray errs = np.PyArray_SimpleNew(2, <np.npy_intp *>I_n.shape + 1, type_num)
    cdef float_t[:, ::1] _errs = errs

    for k in prange(X, schedule='guided', num_threads=num_threads, nogil=True):
        for j in range(Y):
            if W[j, k] > 0.0:
                _errs[j, k] = FVU_interp(I_n, W[j, k], I0, di, dj, j, k,
                                         u[0, j, k], u[1, j, k], ds_y, ds_x, sigma, f)
            else:
                _errs[j, k] = 0.0

    return errs

def pm_total_error(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
                   float_t[:, :, ::1] u not None, float_t[::1] di not None, float_t[::1] dj not None,
                   double ds_y, double ds_x, double sigma, str loss='Huber', unsigned num_threads=1) -> double:
    r"""Return the mean residual for the pixel mapping fit.

    Args:
        I_n (numpy.ndarray) : Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Sample's translations along the vertical detector
            axis in pixels.
        dj (numpy.ndarray) : Sample's translations along the fast detector
            axis in pixels.
        ds_y (float) : Sampling interval of reference image in pixels along
            the vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along
            the horizontal axis.
        sigma (float) : The standard deviation of `I_n`.
        loss (str): Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.
        num_threads (int): Number of threads.

    Returns:
        float : Mean residual value of the pixel mapping fit.

    See Also:
        :func:`pyrost.bin.pm_gsearch` : Description of error metric which
        is being minimized.

    Notes:
        The error metric is given by:

        .. math::

            \varepsilon_{pm}[i, j] = \frac{1}{N}
            \sum_{n = 0}^N f\left( \frac{I[n, i, j] - W[i, j]
            I_{ref}[u[0, i, j] - di[n], u[1, i, j] - dj[n]]}{\sigma}
            \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int Y = I_n.shape[1], X = I_n.shape[2], j, k
    cdef double err

    for k in prange(X, schedule='guided', num_threads=num_threads, nogil=True):
        for j in range(Y):
            if W[j, k] > 0.0:
                err += FVU_interp(I_n, W[j, k], I0, di, dj, j, k,
                                  u[0, j, k], u[1, j, k], ds_y, ds_x, sigma, f)

    return err / (X * Y)

cdef void FVU_frame(float_t[:, ::1] errors, float_t[:, ::1] R, float_t[:, ::1] I0,
                    uint_t[:, ::1] I_n, float_t[:, ::1] W, float_t[:, :, ::1] u, float_t di, float_t dj,
                    double ds_y, double ds_x, double h, double sigma, loss_func f) nogil:
    cdef int Y = I_n.shape[0], X = I_n.shape[1]
    cdef int j, k, jj, kk, j0, k0, jj0, jj1, kk0, kk1
    cdef int Y0 = I0.shape[0], X0 = I0.shape[1]
    cdef int dn_y = <int>ceil((4.0 * h) / ds_y), dn_x = <int>ceil((4.0 * h) / ds_x)
    cdef double y, x, r, I

    for j in range(Y):
        for k in range(X):
            y = u[0, j, k] - di
            x = u[1, j, k] - dj

            j0 = <int>(y / ds_y) + 1
            k0 = <int>(x / ds_x) + 1

            jj0 = j0 - dn_y if j0 - dn_y > 0 else 0
            jj1 = j0 + dn_y if j0 + dn_y < Y0 else Y0
            kk0 = k0 - dn_x if k0 - dn_x > 0 else 0
            kk1 = k0 + dn_x if k0 + dn_x < X0 else X0

            for jj in range(jj0, jj1):
                for kk in range(kk0, kk1):
                    r = rbf((ds_y * jj - y) * (ds_y * jj - y) + (ds_x * kk - x) * (ds_x * kk - x), h)
                    errors[jj, kk] += r * f((<double>I_n[j, k] - W[j, k] * I0[jj, kk]) / sigma)
                    R[jj, kk] += r

def ref_errors(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
               float_t[:, :, ::1] u not None, float_t[::1] di not None, float_t[::1] dj not None,
               double ds_y, double ds_x, double h, double sigma, str loss='Huber', unsigned num_threads=1) -> np.ndarray:
    r"""Return the residuals for the reference image regression.

    Args:
        I_n (numpy.ndarray) :  Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray): Sample's translations along the fast detector
            axis in pixels.
        ds_y (float) : Sampling interval of reference image in pixels along
            the vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along
            the horizontal axis.
        h (float) : Kernel bandwidth in pixels.
        sigma (float) : The standard deviation of `I_n`.
        loss (str) : Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.

        num_threads (int) : Number of threads.

    Returns:
        numpy.ndarray : Residuals array of the reference image regression.

    Notes:
        The pixel mapping `u` maps the intensity measurements from the
        detector plane to the reference plane as follows:

        .. math::
            i_{ref}[n, i, j] = u[0, i, j] + di[n], \; j_{ref}[n, i, j] = u[1, i, j] + dj[n]

        The error metric is given by:

        .. math::

            \varepsilon_{ref}[i, j] = \sum_{n, i^{\prime}, j^{\prime}}
            K[i - i_{ref}[n, i^{\prime}, j^{\prime}], 
            j - j_{ref}[n, i^{\prime}, j^{\prime}], h] \;
            f\left( \frac{I_n[n, i^{\prime}, j^{\prime}] -
            W[i^{\prime}, j^{\prime}] I_{ref}[i, j]}{\sigma} \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function
        and :math:`K[i, j, h] = \frac{1}{\sqrt{2 \pi}} \exp(-\frac{i^2 + j^2}{h})`
        is the Gaussian kernel.

    See Also:
        :func:`pyrost.bin.KR_reference` : Description of the reference image
        estimator.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int N = I_n.shape[0], Y0 = I0.shape[0], X0 = I0.shape[1]
    cdef int i, j, k, t
        
    cdef np.npy_intp *shape = [num_threads, Y0, X0]
    cdef float_t[:, :, ::1] err_buf = np.PyArray_ZEROS(3, shape, type_num, 0)
    cdef float_t[:, :, ::1] R_buf = np.PyArray_ZEROS(3, shape, type_num, 0)
    cdef np.ndarray err = np.PyArray_SimpleNew(2, shape + 1, type_num)

    for i in prange(N, schedule='guided', num_threads=num_threads, nogil=True):
        t = openmp.omp_get_thread_num()
        FVU_frame(err_buf[t], R_buf[t], I0, I_n[i], W, u, di[i], dj[i],
                  ds_y, ds_x, h, sigma, f)

    cdef float_t[:, ::1] _err = err
    cdef float_t err_sum, R_sum
    for k in prange(X0, schedule='guided', num_threads=num_threads, nogil=True):
        for j in range(Y0):
            err_sum = 0.0; R_sum = 0.0
            for i in range(<int>num_threads):
                err_sum = err_sum + err_buf[i, j, k]
                R_sum = R_sum + R_buf[i, j, k]

            _err[j, k] = err_sum / R_sum if R_sum > 0.0 else 0.0

    return err

def ref_total_error(uint_t[:, :, ::1] I_n not None, float_t[:, ::1] W not None, float_t[:, ::1] I0 not None,
                    float_t[:, :, ::1] u not None, float_t[::1] di not None, float_t[::1] dj not None,
                    double ds_y, double ds_x, double h, double sigma, str loss='Huber', unsigned num_threads=1):
    r"""Return the mean residual for the reference image regression.

    Args:
        I_n (numpy.ndarray) :  Measured intensity frames.
        W (numpy.ndarray) : Measured frames' whitefield.
        I0 (numpy.ndarray) : Reference image of the sample.
        u (numpy.ndarray) : The discrete geometrical mapping of the detector
            plane to the reference image.
        di (numpy.ndarray) : Sample's translations along the vertical
            detector axis in pixels.
        dj (numpy.ndarray): Sample's translations along the fast detector
            axis in pixels.
        ds_y (float) : Sampling interval of reference image in pixels along
            the vertical axis.
        ds_x (float) : Sampling interval of reference image in pixels along
            the horizontal axis.
        h (float) : Kernel bandwidth in pixels.
        sigma (float) : The standard deviation of `I_n`.
        loss (str) : Choose between the following loss functions:

            * 'Epsilon': Epsilon loss function (epsilon = 0.5)
            * 'Huber' : Huber loss function (k = 1.345)
            * 'L1' : L1 norm loss function.
            * 'L2' : L2 norm loss function.

        num_threads (int) : Number of threads.

    Returns:
        float : Mean residual value.

    Notes:
        The pixel mapping `u` maps the intensity measurements from the
        detector plane to the reference plane as follows:

        .. math::
            i_{ref}[n, i, j] = u[0, i, j] + di[n], \; j_{ref}[n, i, j] = u[1, i, j] + dj[n]

        The error metric is given by:

        .. math::

            \varepsilon_{ref}[i, j] = \sum_{n, i^{\prime}, j^{\prime}}
            K[i - i_{ref}[n, i^{\prime}, j^{\prime}], 
            j - j_{ref}[n, i^{\prime}, j^{\prime}], h] \;
            f\left( \frac{I_n[n, i^{\prime}, j^{\prime}] -
            W[i^{\prime}, j^{\prime}] I_{ref}[i, j]}{\sigma} \right)

        where :math:`f(x)` is L1 norm, L2 norm or Huber loss function
        and :math:`K[i, j, h] = \frac{1}{\sqrt{2 \pi}} \exp(-\frac{i^2 + j^2}{h})`
        is the Gaussian kernel.

    See Also:
        :func:`pyrost.bin.KR_reference` : Description of the reference image
        estimator.
    """
    if ds_y <= 0.0 or ds_x <= 0.0:
        raise ValueError('Sampling intervals must be positive')

    cdef loss_func f = choose_loss(loss)

    cdef int type_num = np.PyArray_TYPE(W.base)
    cdef int N = I_n.shape[0], Y0 = I0.shape[0], X0 = I0.shape[1]
    cdef int i, j, t, k
        
    cdef np.npy_intp *shape = [num_threads, Y0, X0]
    cdef float_t[:, :, ::1] err_buf = np.PyArray_ZEROS(3, shape, type_num, 0)
    cdef float_t[:, :, ::1] R_buf = np.PyArray_ZEROS(3, shape, type_num, 0)

    for i in prange(N, schedule='guided', num_threads=num_threads, nogil=True):
        t = openmp.omp_get_thread_num()
        FVU_frame(err_buf[t], R_buf[t], I0, I_n[i], W, u, di[i], dj[i],
                  ds_y, ds_x, h, sigma, f)

    cdef double err = 0.0, err2 = 0.0
    cdef double err_sum, R_sum, err_jk
    for k in prange(X0, schedule='guided', num_threads=num_threads, nogil=True):
        for j in range(Y0):
            err_sum = 0.0; R_sum = 0.0
            for i in range(<int>num_threads):
                err_sum = err_sum + err_buf[i, j, k]
                R_sum = R_sum + R_buf[i, j, k]

            if R_sum > 0.0:
                err_jk = err_sum / R_sum
                err += err_jk
                err2 += err_jk * err_jk

    err /= (X0 * Y0)
    err2 /= (X0 * Y0)

    return err, sqrt(err2 - err * err)

def ct_integrate(float_t[:, ::1] sy_arr not None, float_t[:, ::1] sx_arr not None, int num_threads=1) -> np.ndarray:
    """Perform the Fourier Transform wavefront reconstruction [FTI]_ with
    antisymmetric derivative integration [ASDI]_.

    Args:
        sx_arr (numpy.ndarray) :  of gradient values along the horizontal axis.
        sy_arr (numpy.ndarray) : Array of gradient values along the vertical axis.
        num_threads (int) : Number of threads.

    Returns:
        numpy.ndarray : Reconstructed wavefront.

    References:
        .. [FTI] C. Kottler, C. David, F. Pfeiffer, and O. Bunk,
                "A two-directional approach for grating based
                differential phase contrast imaging using hard x-rays,"
                Opt. Express 15, 1175-1181 (2007).
        .. [ASDI] Pierre Bon, Serge Monneret, and Benoit Wattellier,
                "Noniterative boundary-artifact-free wavefront
                reconstruction from its derivatives," Appl. Opt. 51,
                5698-5704 (2012).
    """
    cdef int type_num = np.PyArray_TYPE(sx_arr.base)
    cdef np.npy_intp a = sx_arr.shape[0], b = sx_arr.shape[1]
    cdef int i, j, ii, jj
    cdef np.npy_intp *asdi_shape = [2 * a, 2 * b]
    
    cdef np.ndarray[np.complex128_t, ndim=2] sfy_asdi = np.PyArray_SimpleNew(2, asdi_shape, np.NPY_COMPLEX128)
    cdef pyfftw.FFTW fftw_obj = pyfftw.FFTW(sfy_asdi, sfy_asdi, axes=(0, 1), threads=num_threads)
    for i in range(a):
        for j in range(b):
            sfy_asdi[i, j] = -sy_arr[a - i - 1, b - j - 1]
    for i in range(a):
        for j in range(b):
            sfy_asdi[i + a, j] = sy_arr[i, b - j - 1]
    for i in range(a):
        for j in range(b):
            sfy_asdi[i, j + b] = -sy_arr[a - i - 1, j]
    for i in range(a):
        for j in range(b):
            sfy_asdi[i + a, j + b] = sy_arr[i, j]
    fftw_obj._execute()

    cdef np.ndarray[np.complex128_t, ndim=2] sfx_asdi = np.PyArray_SimpleNew(2, asdi_shape, np.NPY_COMPLEX128)
    fftw_obj._update_arrays(sfx_asdi, sfx_asdi)
    for i in range(a):
        for j in range(b):
            sfx_asdi[i, j] = -sx_arr[a - i - 1, b - j - 1]
    for i in range(a):
        for j in range(b):
            sfx_asdi[i + a, j] = -sx_arr[i, b - j - 1]
    for i in range(a):
        for j in range(b):
            sfx_asdi[i, j + b] = sx_arr[a - i - 1, j]
    for i in range(a):
        for j in range(b):
            sfx_asdi[i + a, j + b] = sx_arr[i, j]
    fftw_obj._execute()

    cdef pyfftw.FFTW ifftw_obj = pyfftw.FFTW(sfx_asdi, sfx_asdi, direction='FFTW_BACKWARD', axes=(0, 1), threads=num_threads)
    cdef double xf, yf, norm = 1.0 / <double>np.PyArray_SIZE(sfx_asdi)
    for i in range(2 * a):
        yf = 0.5 * <double>i / a - i // a
        for j in range(2 * b):
            xf = 0.5 * <double>j / b - j // b
            sfx_asdi[i, j] = norm * (sfy_asdi[i, j] * yf + sfx_asdi[i, j] * xf) / (2j * pi * (xf * xf + yf * yf))
    sfx_asdi[0, 0] = 0.0 + 0.0j
    ifftw_obj._execute()

    return np.asarray(sfx_asdi.real[a:, b:], dtype=sx_arr.base.dtype)


cdef extern from "gsl/gsl_rng.h":

    ctypedef struct gsl_rng_type
    ctypedef struct gsl_rng

    cdef gsl_rng_type *gsl_rng_mt19937

    gsl_rng *gsl_rng_alloc(gsl_rng_type * T) nogil

    unsigned long int gsl_rng_get(gsl_rng * r) nogil

    void gsl_rng_set(gsl_rng * r, unsigned long int seed) nogil

    void gsl_rng_free(gsl_rng * r) nogil

    double gsl_rng_uniform(gsl_rng * r) nogil
    unsigned long gsl_rng_uniform_int(gsl_rng * r, unsigned long n) nogil


"""
bin - cython functions package
"""
from __future__ import absolute_import

from .simulation import (next_fast_len, gaussian_kernel, gaussian_filter,
                         gaussian_gradient_magnitude, rsc_wp, fraunhofer_wp,
                         bar_positions, barcode_profile, mll_profile,
                         fft_convolve, make_frames, median, median_filter)
from .pyrost import (KR_reference, LOWESS_reference, pm_gsearch, pm_rsearch,
                     pm_devolution, tr_gsearch, pm_errors, pm_total_error,
                     ref_errors, ref_total_error, ct_integrate)
from .pyfftw import FFTW, empty_aligned, zeros_aligned, ones_aligned



"""CXI file loader class (:class:`pyrost.CXILoader`) uses a protocol to
automatically load all the necessary data fields from a CXI file.

:class:`pyrost.STData` contains all the necessary data for the Speckle
Tracking algorithm, and provides a suite of data processing tools to work
with the data.

Examples:
    Extract all the necessary data using a :func:`pyrost.cxi_loader` function.

    >>> import pyrost as rst
    >>> loader = rst.cxi_loader()
    >>> rst_data = loader.load('results/test/data.cxi')

    Mask the bad pixels, integrate the measured frames along the vertical axis,
    mirror the data, and crop it using a region of interest as follows:

    >>> data = data.update_mask(pmax=99.999, update='multiply')
    >>> data = data.integrate_data(axis=0)
    >>> data = data.crop_data(roi=(0, 1, 200, 1240))
    >>> data = data.mirror_data(axis=0)
"""
from __future__ import annotations
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
from weakref import ref
from multiprocessing import cpu_count
from tqdm.auto import tqdm
import h5py
import numpy as np
from .aberrations_fit import AberrationsFit
from .data_container import DataContainer, dict_to_object
from .cxi_protocol import CXIProtocol, CXI_PROTOCOL
from .rst_update import SpeckleTracking
from .bin import median, median_filter, fft_convolve, ct_integrate

class CXILoader(CXIProtocol):
    """CXI file loader class. Loads data from a CXI file and returns
    a :class:`STData` container or a :class:`dict` with the data. Search
    data in the paths provided by `protocol` and `load_paths`.

    Attributes:
        config : Protocol configuration.
        datatypes : Dictionary with attributes' datatypes. 'float', 'int',
            or 'bool' are allowed.
        default_paths : Dictionary with attributes' CXI default file paths.
        is_data : Dictionary with the flags if the attribute is of data
            type. Data type is 2- or 3-dimensional and has the same data
            shape as `data`.
        load_paths : Extra set of paths to the attributes enlisted in
            `datatypes`.
        policy: Loading policy. If a flag for the given attribute is
            True, the attribute will be loaded from a file.

    See Also:
        :class:`pyrost.STData` : Data container with all the data  necessary
            for Speckle Tracking.
    """
    attr_dict = {'config': ('float_precision', ), 'datatypes': ('ALL', ),
                 'default_paths': ('ALL', ), 'load_paths': ('ALL',),
                 'is_data': ('ALL', ), 'policy': ('ALL', )}
    fmt_dict = {'config': 'str', 'datatypes': 'str', 'default_paths': 'str',
                'load_paths': 'str', 'is_data': 'str', 'policy': 'str'}

    def __init__(self, protocol: CXIProtocol, load_paths: Dict[str, List[str]],
                 policy: Dict[str, Union[str, bool]]) -> None:
        """
        Args:
            protocol : Protocol object.
            load_paths : Extra paths to the data attributes in a CXI file,
                which override `protocol`. Accepts only the attributes
                enlisted in `protocol`.
            policy : A dictionary with loading policy. Contains all the
                attributes that are available in `protocol` with their
                corresponding flags. If a flag is True, the attribute
                will be loaded from a file.
        """
        load_paths = {attr: paths for attr, paths in load_paths.items()
                      if attr in protocol}
        policy = {attr: flag for attr, flag in policy.items() if attr in protocol}
        super(CXIProtocol, self).__init__(config=protocol.config, datatypes=protocol.datatypes,
                                          default_paths=protocol.default_paths,
                                          is_data=protocol.is_data, load_paths=load_paths,
                                          policy=policy)

        if self.config['float_precision'] == 'float32':
            self.known_types['float'] = np.float32
        elif self.config['float_precision'] == 'float64':
            self.known_types['float'] = np.float64
        else:
            raise ValueError('Invalid float precision: {:s}'.format(self.config['float_precision']))

    @staticmethod
    def str_to_list(strings: Union[str, List[str]]) -> List[str]:
        """Convert `strings` to a list of strings.

        Args:
            strings : String or a list of strings

        Returns:
            List of strings.
        """
        if isinstance(strings, (str, list)):
            if isinstance(strings, str):
                return [strings,]
            return strings

        raise ValueError('strings must be a string or a list of strings')

    @classmethod
    def import_default(cls, protocol: Optional[CXIProtocol]=None,
                       load_paths: Optional[Dict[str, List[str]]]=None,
                       policy: Optional[Dict[str, Union[str, bool]]]=None) -> CXILoader:
        """Return the default :class:`CXILoader` object. Extra arguments
        override the default values if provided.

        Args:
            protocol : Protocol object.
            load_paths : Extra paths to the data attributes in a CXI file,
                which override `protocol`. Accepts only the attributes
                enlisted in `protocol`.
            policy : A dictionary with loading policy. Contains all the
                attributes that are available in `protocol` and the
                corresponding flags. If a flag is True, the attribute
                will be loaded from a file.

        Returns:
            A :class:`CXILoader` object with the default parameters.
        """
        return cls.import_ini(CXI_PROTOCOL, protocol, load_paths, policy)

    @classmethod
    def import_ini(cls, ini_file: str, protocol: Optional[CXIProtocol]=None,
                   load_paths: Optional[Dict[str, List[str]]]=None,
                   policy: Optional[Dict[str, Union[str, bool]]]=None) -> CXILoader:
        """Initialize a :class:`CXILoader` object class with an
        ini file.

        Args:
            ini_file : Path to the ini file. Loads the default CXI loader if None.
            protocol : Protocol object. Initialized with `ini_file` if None.
            load_paths : Extra paths to the data attributes in a CXI file,
                which override `protocol`. Accepts only the attributes
                enlisted in `protocol`. Initialized with `ini_file`
                if None.
            policy : A dictionary with loading policy. Contains all the
                attributes that are available in `protocol` and the
                corresponding flags. If a flag is True, the attribute
                will be loaded from a file. Initialized with `ini_file`
                if None.

        Returns:
            A :class:`CXILoader` object with all the attributes imported
            from the ini file.
        """
        if protocol is None:
            protocol = CXIProtocol.import_ini(ini_file)
        kwargs = cls._import_ini(ini_file)
        if not load_paths is None:
            kwargs['load_paths'].update(**load_paths)
        if not policy is None:
            kwargs['policy'].update(**policy)
        return cls(protocol=protocol, load_paths=kwargs['load_paths'],
                   policy=kwargs['policy'])

    def get_load_paths(self, attr: str, value: Optional[Union[str, List[str]]]=None) -> List[str]:
        """Return the atrribute's path in the cxi file.
        Return `value` if `attr` is not found.

        Args:
            attr : The attribute to look for.
            value : Value which is returned if the `attr` is not
                found.

        Returns:
            Set of attribute's paths.
        """
        paths = self.str_to_list(super(CXILoader, self).get_default_path(attr, value))
        if attr in self.load_paths:
            paths.extend(self.load_paths[attr])
        return paths

    def get_policy(self, attr: str, value: bool=False) -> bool:
        """Return the atrribute's loding policy.

        Args:
            attr : The attribute to look for.
            value : Value which is returned if the `attr` is not
                found.

        Returns:
            Attributes' loding policy.
        """
        policy = self.policy.get(attr, value)
        if isinstance(policy, str):
            return policy in ['True', 'true', '1', 'y', 'yes']
        else:
            return bool(policy)

    def get_protocol(self) -> CXIProtocol:
        """Return a CXI protocol from the loader.

        Returns:
            CXI protocol.
        """
        return CXIProtocol(datatypes=self.datatypes,
                           default_paths=self.default_paths,
                           is_data=self.is_data,
                           float_precision=self.config['float_precision'])

    def find_path(self, attr: str, cxi_file: h5py.File) -> str:
        """Find attribute's path in a CXI file `cxi_file`.

        Args:
            attr : Data attribute.
            cxi_file : :class:`h5py.File` object of the CXI file.

        Returns:
            Atrribute's path in the CXI file, returns an empty
            string if the attribute is not found.
        """
        paths = self.get_load_paths(attr)
        for path in paths:
            if path in cxi_file:
                return path
        return str()

    def load_attributes(self, master_file: str) -> Dict[str, Any]:
        """Return attributes' values from a CXI file at
        the given `master_file`.

        Args:
            master_file : Path to the master CXI file.

        Returns:
            Dictionary with the attributes retrieved from
            the CXI file.
        """
        attr_dict = {}
        with h5py.File(master_file, 'r') as cxi_file:
            for attr in self:
                cxi_path = self.find_path(attr, cxi_file)
                if not self.get_is_data(attr) and self.get_policy(attr, False) and cxi_path:
                    attr_dict[attr] = self.read_cxi(attr, cxi_file, cxi_path)
        return attr_dict

    def read_indices(self, attr: str, data_files: Union[str, List[str]]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Retrieve the indices of the datasets from the CXI files for the
        given attribute `attr`.

        Args:
            attr : The attribute to read.
            data_files : Paths to the data CXI files.

        Returns:
            A tuple of ('paths', 'cxi_paths', 'indices'). The elements
            are the following:

            * 'paths' : List of the file paths. None if no data is found.
            * 'cxi_paths' : List of the paths inside the files. None if no
              data is found.
            * 'indices' : List of the frame indices. None if no data is found.

        """
        data_files = self.str_to_list(data_files)

        paths, cxi_paths, indices = [], [], []
        for path in data_files:
            with h5py.File(path, 'r') as cxi_file:
                shapes = self.read_shape(cxi_file, self.find_path(attr, cxi_file))
            if shapes:
                for cxi_path, dset_shape in shapes:
                    if len(dset_shape) == 3:
                        paths.append(np.repeat(path, dset_shape[0]))
                        cxi_paths.append(np.repeat(cxi_path, dset_shape[0]))
                        indices.append(np.arange(dset_shape[0]))
                    elif len(dset_shape) == 2:
                        paths.append(np.atleast_1d(path))
                        cxi_paths.append(np.atleast_1d(cxi_path))
                        indices.append(np.atleast_1d(slice(None)))
                    else:
                        raise ValueError('Dataset must be 2- or 3-dimensional')

        if len(paths) == 0:
            return np.array([]), np.array([]), np.array([])
        elif len(paths) == 1:
            return paths[0], cxi_paths[0], indices[0]
        else:
            return (np.concatenate(paths), np.concatenate(cxi_paths),
                    np.concatenate(indices))

    def load_data(self, attr: str, paths: np.ndarray, cxi_paths: np.ndarray, indices: np.ndarray,
                  verbose: bool=True) -> np.ndarray:
        """Retrieve the data for the given attribute `attr` from the
        CXI files. Uses the result from :func:`CXILoader.read_indices`
        method.

        Args:
            attr : The attribute to read.
            paths : List of the file paths.
            cxi_paths : List of the paths inside the files.
            indices : List of the frame indices.
            verbose : Print the progress bar if True.

        Returns:
            Data array retrieved from the CXI files.
        """
        data = []
        for path, cxi_path, index in tqdm(zip(paths, cxi_paths, indices),
                                          disable=not verbose, total=paths.size,
                                          desc=f'Loading {attr:s}'):
            with h5py.File(path, 'r') as cxi_file:
                data.append(cxi_file[cxi_path][index])

        if len(data) == 1:
            data = data[0]
        else:
            data = np.stack(data, axis=0)
        return np.asarray(data, dtype=self.get_dtype(attr))

    def load_to_dict(self, data_files: Union[str, List[str]],
                     master_file: Optional[str]=None,
                     frame_indices: Optional[Iterable[int]]=None,
                     **attributes: Any) -> Dict[str, Any]:
        """Load data from the CXI files and return a :class:`dict` with
        all the data fetched from the `data_files` and `master_file`.

        Args:
            data_files : Paths to the data CXI files.
            master_file : Path to the master CXI file. First file in `data_files`
                if not provided.
            frame_indices : Array of frame indices to load. Loads all the frames
                by default.
            attributes : Dictionary of attribute values, that override the loaded
                values.

        Returns:
            Dictionary with all the data fetched from the CXI files.
        """
        if master_file is None:
            if isinstance(data_files, str):
                master_file = data_files
            elif isinstance(data_files, list):
                master_file = data_files[0]
            else:
                raise ValueError('data_files must be a string or a list of strings')

        data_dict = self.load_attributes(master_file)

        if frame_indices is None:
            n_frames = 0
            for attr in self:
                if self.get_is_data(attr) and self.get_policy(attr):
                    n_frames = max(self.read_indices(attr, data_files)[0].size, n_frames)
            frame_indices = np.arange(n_frames)
        else:
            frame_indices = np.asarray(frame_indices)

        if frame_indices.size:
            for attr in self:
                if self.get_is_data(attr) and self.get_policy(attr):
                    paths, cxi_paths, indices = self.read_indices(attr, data_files)
                    if paths.size > 0:
                        good_frames = frame_indices[frame_indices < paths.size]
                        data_dict[attr] = self.load_data(attr, paths[good_frames],
                                                         cxi_paths[good_frames],
                                                         indices[good_frames])

        for attr, val in attributes.items():
            if attr in self and val is not None:
                if isinstance(val, dict):
                    data_dict[attr] = {dkey: self.get_dtype(attr)(dval)
                                       for dkey, dval in val.items()}
                else:
                    data_dict[attr] = np.asarray(val, dtype=self.get_dtype(attr))
                    if data_dict[attr].size == 1:
                        data_dict[attr] = data_dict[attr].item()

        if 'mask' in data_dict and 'data' in data_dict:
            if data_dict['mask'].shape == data_dict['data'].shape[1:]:
                data_dict['mask'] = np.tile(data_dict['mask'][None, :],
                                            (data_dict['data'].shape[0], 1, 1))

        return data_dict

    def load(self, data_files: str, master_file: Optional[str]=None,
             frame_indices: Optional[Iterable[str]]=None, **attributes: Any) -> STData:
        """Load data from the CXI files and return a :class:`STData` container
        with all the data fetched from the `data_files` and `master_file`.

        Args:
            data_files : Paths to the data CXI files.
            master_file : Path to the master CXI file. First file in `data_files`
                if not provided.
            frame_indices : Array of frame indices to load. Loads all the frames by
                default.
            attributes : Dictionary of attribute values, which will be parsed to
                the :class:`pyrost.STData` object instead.

        Returns:
            Data container object with all the necessary data for the speckle
            tracking algorithm.
        """
        return STData(self.get_protocol(), **self.load_to_dict(data_files, master_file,
                                                               frame_indices, **attributes))

class STData(DataContainer):
    """Speckle tracking data container class. Contains all the necessary data
    for the robust speckle tracking data processing pipeline (list of data
    attributes is specified in `attr_set` and `init_set`) and provides a suite
    of data processing tools to work with the data.

    Attributes:
        attr_set : Set of attributes in the container which are necessary
            to initialize in the constructor.
        init_set : Set of optional data attributes.

    Notes:
        **Necessary attributes**:

        * basis_vectors : Detector basis vectors
        * data : Measured intensity frames.
        * distance : Sample-to-detector distance [m].
        * translations : Sample's translations [m].
        * wavelength : Incoming beam's wavelength [m].
        * x_pixel_size : Pixel's size along the horizontal detector axis [m].
        * y_pixel_size : Pixel's size along the vertical detector axis [m].

        **Optional attributes**:

        * defocus_x : Defocus distance for the horizontal detector axis [m].
        * defocus_y : Defocus distance for the vertical detector axis [m].
        * error_frame : MSE (mean-squared-error) of the reference image
          and pixel mapping fit per pixel.
        * flatfields : Set of flatfields for each of the measured images.
        * good_frames : An array of good frames' indices.
        * mask : Bad pixels mask.
        * num_threads : Number of threads used in computations.
        * phase : Phase profile of lens' aberrations.
        * pixel_aberrations : Lens' aberrations along the horizontal and
          vertical axes in pixels.
        * pixel_map : The pixel mapping between the data at the detector's
          plane and the reference image at the reference plane.
        * pixel_translations : Sample's translations in the detector's
          plane in pixels.
        * reference_image : The unabberated reference image of the sample.
        * roi : Region of interest in the detector plane.
        * sigma : Standard deviation of measured frames.
        * whitefield : Measured frames' whitefield.
    """
    attr_set = {'protocol', 'basis_vectors', 'data', 'distance', 'translations', 'wavelength',
                'x_pixel_size', 'y_pixel_size'}
    init_set = {'defocus_x', 'defocus_y', 'error_frame', 'flatfields', 'good_frames',
                'mask', 'num_threads', 'phase', 'pixel_aberrations', 'pixel_map',
                'pixel_translations', 'reference_image', 'roi', 'sigma', 'whitefield'}

    inits = {'num_threads': lambda obj: np.clip(1, 64, cpu_count()),
             'roi'        : lambda obj: np.array([0, obj.data.shape[1], 0, obj.data.shape[2]]),
             'good_frames': lambda obj: np.arange(obj.data.shape[0]),
             'mask'       : lambda obj: np.ones(obj.data.shape, dtype=bool),
             'whitefield' : lambda obj: median(data=obj.data[obj.good_frames], axis=0,
                                               mask=obj.mask[obj.good_frames],
                                               num_threads=obj.num_threads),
             'sigma'      : lambda obj: np.std(obj.data[:, obj.roi[0]:obj.roi[1],
                                                        obj.roi[2]:obj.roi[3]] * \
                                               obj.mask[:, obj.roi[0]:obj.roi[1],
                                                        obj.roi[2]:obj.roi[3]]),
             'defocus_y'  : lambda obj: obj.get('defocus_x', None),
             'pixel_map'  : lambda obj: obj._pixel_map(),
             'pixel_translations': lambda obj: obj._pixel_translations()}

    def __init__(self, protocol: CXIProtocol=CXIProtocol.import_default(),
                 **kwargs: Union[int, float, np.ndarray]) -> None:
        """
        Args:
            protocol : CXI protocol.
            kwargs : Dictionary of the necessary and optional data attributes specified
                in :class:`pyrost.STData` notes. All the necessary attributes must be
                provided

        Raises:
            ValueError : If any of the necessary attributes specified in :class:`pyrost.STData`
            notes have not been provided.
        """
        super(STData, self).__init__(protocol=protocol, **kwargs)

    @property
    def _isdefocus(self) -> bool:
        return self.defocus_x is not None

    @property
    def _isphase(self) -> bool:
        return not self.pixel_aberrations is None and not self.phase is None

    def _pixel_translations(self) -> Union[np.ndarray, None]:
        if not self._isdefocus:
            return None

        pixel_translations = (self.translations[:, None] * self.basis_vectors).sum(axis=-1)
        mag = np.abs(self.distance / np.array([self.defocus_y, self.defocus_x]))
        pixel_translations *= mag / (self.basis_vectors**2).sum(axis=-1)
        pixel_translations -= pixel_translations[0]
        pixel_translations -= pixel_translations.mean(axis=0)
        return pixel_translations

    def _pixel_map(self) -> np.ndarray:
        pixel_map = np.indices(self.whitefield.shape, dtype=float)
        if self._isdefocus:
            if self.defocus_y < 0.0 and pixel_map[0, 0, 0] < pixel_map[0, -1, 0]:
                pixel_map = np.flip(pixel_map, axis=1)
            if self.defocus_x < 0.0 and pixel_map[1, 0, 0] < pixel_map[1, 0, -1]:
                pixel_map = np.flip(pixel_map, axis=2)
        return pixel_map

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in self and 'protocol' in self.__dict__:
            dtype = self.protocol.get_dtype(attr)
            if isinstance(value, np.ndarray):
                value = np.asarray(value, dtype=dtype)
            super(STData, self).__setattr__(attr, value)
        else:
            super(STData, self).__setattr__(attr, value)

    @dict_to_object
    def bin_data(self, bin_ratio: int=2) -> STData:
        """Return a new :class:`STData` object with the data binned by
        a factor `bin_ratio`.

        Args:
            bin_ratio : Binning ratio. The frame size will decrease by
                the factor of `bin_ratio`.

        Returns:
            New :class:`STData` object with binned `data`.
        """
        data_dict = {'basis_vectors': bin_ratio * self.basis_vectors,
                     'roi': self.roi // bin_ratio + (self.roi % bin_ratio > 0),
                     'pixel_map': self.pixel_map / 2,
                     'x_pixel_size': bin_ratio * self.x_pixel_size,
                     'y_pixel_size': bin_ratio * self.y_pixel_size}
        if self._isdefocus:
            data_dict['pixel_translations'] = self.pixel_translations / bin_ratio

        for attr, val in self.items():
            if val is not None and self.protocol.get_is_data(attr):
                if attr in data_dict:
                    data_dict[attr] = data_dict[attr][..., ::bin_ratio, ::bin_ratio]
                else:
                    data_dict[attr] = val[..., ::bin_ratio, ::bin_ratio]
        return data_dict

    @dict_to_object
    def crop_data(self, roi: Iterable[int]) -> STData:
        """Return a new :class:`STData` object with the updated `roi`.

        Args:
            roi : Region of interest in the detector plane.

        Returns:
            New :class:`STData` object with the updated `roi`.
        """
        return {'roi': np.asarray(roi, dtype=int), 'flatfields': None, 'sigma': None}

    @dict_to_object
    def integrate_data(self, axis: int=0) -> STData:
        """Return a new :class:`STData` object with the `data` summed
        over the `axis`.

        Args:
            axis : Axis along which a sum is performed.

        Returns:
            New :class:`STData` object with the stack of measured
            frames integrated along the given axis.
        """
        roi = self.roi.copy()
        roi[2 * axis:2 * (axis + 1)] = np.arange(2)

        data = np.zeros(self.data.shape, self.data.dtype)
        data[self.good_frames, self.roi[0]:self.roi[1],
             self.roi[2]:self.roi[3]] = self.get('data') * self.get('mask')
        return {'data': np.sum(data, axis=axis + 1, keepdims=True),
                'flatfields': None, 'mask': None, 'pixel_map': None,
                'pixel_translations': None, 'roi': roi, 'sigma': None,
                'whitefield': None}

    @dict_to_object
    def mask_frames(self, good_frames: Optional[Iterable[int]]=None) -> STData:
        """Return a new :class:`STData` object with the updated
        good frames mask. Mask empty frames by default.

        Args:
            good_frames : List of good frames' indices. Masks empty
                frames if not provided.

        Returns:
            New :class:`STData` object with the updated `good_frames`
            and `whitefield`.
        """
        if good_frames is None:
            good_frames = np.where(self.data.sum(axis=(1, 2)) > 0)[0]
        return {'good_frames': np.asarray(good_frames, dtype=np.int),
                'whitefield': None}

    @dict_to_object
    def mirror_data(self, axis: int=1) -> STData:
        """Return a new :class:`STData` object with the data mirrored
        along the given axis.

        Args:
            axis : Choose between the vertical axis (0) and
                the horizontal axis (1).

        Returns:
            New :class:`STData` object with the updated `data` and
            `basis_vectors`.
        """
        if axis not in [0, 1]:
            raise ValueError('Axis must equal to 0 or 1')

        basis_vectors = np.copy(self.basis_vectors)
        basis_vectors[:, axis] *= -1.0

        roi = np.copy(self.roi)
        roi[2 * axis] = self.whitefield.shape[axis] - self.roi[2 * axis + 1]
        roi[2 * axis + 1] = self.whitefield.shape[axis] - self.roi[2 * axis]
        data_dict = {'basis_vectors': basis_vectors, 'roi': roi,
                     'pixel_translations': None}

        for attr, val in self.items():
            if val is not None and self.protocol.get_is_data(attr):
                data_dict[attr] = np.flip(val, axis=axis - 2)
        return data_dict

    @dict_to_object
    def update_mask(self, method: str='perc-bad', pmin: float=0., pmax: float=99.99,
                    vmin: int=0, vmax: int=65535, update: str='reset') -> STData:
        """Return a new :class:`STData` object with the updated
        bad pixels mask.

        Args:
            method : Bad pixels masking methods:

                * 'no-bad' (default) : No bad pixels.
                * 'range-bad' : Mask the pixels which values lie outside
                  of (`vmin`, `vmax`) range.
                * 'perc-bad' : Mask the pixels which values lie outside
                  of the (`pmin`, `pmax`) percentiles.

            vmin : Lower intensity bound of 'range-bad' masking method.
            vmax : Upper intensity bound of 'range-bad' masking method.
            pmin : Lower percentage bound of 'perc-bad' masking method.
            pmax : Upper percentage bound of 'perc-bad' masking method.
            update : Multiply the new mask and the old one if 'multiply',
                use the new one if 'reset'.

        Returns:
            New :class:`STData` object with the updated `mask`.
        """
        data = self.get('data')
        if method == 'no-bad':
            mask = np.ones((self.good_frames.size, self.roi[1] - self.roi[0],
                            self.roi[3] - self.roi[2]), dtype=bool)
        elif method == 'range-bad':
            mask = (data >= vmin) & (data < vmax)
        elif method == 'perc-bad':
            offsets = (data - np.median(data))
            mask = (offsets >= np.percentile(offsets, pmin)) & \
                   (offsets <= np.percentile(offsets, pmax))
        else:
            ValueError('Invalid method keyword')

        mask_full = self.mask.copy()
        if update == 'reset':
            mask_full[self.good_frames, self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]] = mask
        elif update == 'multiply':
            mask_full[self.good_frames, self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]] *= mask
        else:
            raise ValueError('Invalid update keyword')

        return {'mask': mask_full, 'sigma': None, 'whitefield': None}

    @dict_to_object
    def update_whitefield(self) -> STData:
        """Return a new :class:`STData` object with the updated `whitefield`.

        Returns:
            New :class:`STData` object with the updated `whitefield`.
        """
        return {'whitefield': None}

    @dict_to_object
    def update_defocus(self, defocus_x: float, defocus_y: Optional[float]=None) -> STData:
        """Return a new :class:`STData` object with the updated defocus
        distances `defocus_x` and `defocus_y` for the horizontal and
        vertical detector axes accordingly. Update `pixel_translations`
        based on the new defocus distances.

        Args:
            defocus_x : Defocus distance for the horizontal detector axis [m].
            defocus_y : Defocus distance for the vertical detector axis [m].
                Equals to `defocus_x` if it's not provided.

        Returns:
            New :class:`STData` object with the updated `defocus_y`,
            `defocus_x`, and `pixel_translations`.
        """
        if defocus_y is None:
            defocus_y = defocus_x
        return {'defocus_y': defocus_y, 'defocus_x': defocus_x,
                'pixel_map': None, 'pixel_translations': None}

    def update_phase(self, st_obj: SpeckleTracking) -> None:
        """Update `pixel_aberrations`, `phase`, and `reference_image`
        based on the data from `st_obj` object. `st_obj` must be derived
        from this data container, an error is raised otherwise.

        Args:
            st_obj : :class:`SpeckleTracking` object derived from this
                data container.

        Raises:
            ValueError : If `st_obj` wasn't derived from this data container.
        """
        if st_obj.parent() is not self:
            raise ValueError("'st_obj' wasn't derived from this data container")
        # Update phase, pixel_aberrations, and reference_image
        dpm_y, dpm_x = (st_obj.pixel_map - self.get('pixel_map'))
        dpm_y -= dpm_y.mean()
        dpm_x -= dpm_x.mean()
        self.pixel_aberrations = np.zeros(self.pixel_map.shape, dtype=self.pixel_map.dtype)
        self.pixel_aberrations[:, self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]] = np.stack((dpm_y, dpm_x))

        # Calculate magnification for horizontal and vertical axes
        mag_y = np.abs((self.distance + self.defocus_y) / self.defocus_y)
        mag_x = np.abs((self.distance + self.defocus_x) / self.defocus_x)

        # Calculate the distance between the reference and the detector plane
        dist_y = self.distance * (1 - mag_y**-1)
        dist_x = self.distance * (1 - mag_x**-1)

        # dTheta = delta_pix / distance / magnification * du
        # Phase = 2 * pi / wavelength * Integrate[dTheta, delta_pix]
        phase = ct_integrate(sy_arr=self.y_pixel_size**2 / dist_y / mag_y * dpm_y,
                             sx_arr=self.x_pixel_size**2 / dist_x / mag_x * dpm_x)
        phase *= 2.0 * np.pi / self.wavelength
        self.phase = np.zeros(self.whitefield.shape, dtype=self.pixel_map.dtype)
        self.phase[self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]] = phase
        self.reference_image = st_obj.reference_image

    def fit_phase(self, center: int=0, axis: int=1, max_order: int=2, xtol: float=1e-14,
                  ftol: float=1e-14, loss: str='cauchy') -> Dict[str, Union[float, np.ndarray]]:
        """Fit `pixel_aberrations` with the polynomial function using nonlinear
        least-squares algorithm. The function uses least-squares algorithm from
        :func:`scipy.optimize.least_squares`.

        Args:
            center : Index of the zerro scattering angle or direct beam pixel.
            axis : Axis along which `pixel_aberrations` is fitted.
            max_order : Maximum order of the polynomial model function.
            xtol : Tolerance for termination by the change of the independent
                variables.
            ftol : Tolerance for termination by the change of the cost function.
            loss : Determines the loss function. The following keyword values are
                allowed:

                * 'linear' : ``rho(z) = z``. Gives a standard
                  least-squares problem.
                * 'soft_l1' : ``rho(z) = 2 * ((1 + z)**0.5 - 1)``. The smooth
                  approximation of l1 (absolute value) loss. Usually a good
                  choice for robust least squares.
                * 'huber' : ``rho(z) = z if z <= 1 else 2*z**0.5 - 1``. Works
                  similarly to 'soft_l1'.
                * 'cauchy' (default) : ``rho(z) = ln(1 + z)``. Severely weakens
                  outliers influence, but may cause difficulties in optimization
                  process.
                * 'arctan' : ``rho(z) = arctan(z)``. Limits a maximum loss on
                  a single residual, has properties similar to 'cauchy'.

        Returns:
            A dictionary with the model fit information. The following fields
            are contained:

            * 'c_3' : Third order aberrations coefficient [rad / mrad^3].
            * 'c_4' : Fourth order aberrations coefficient [rad / mrad^4].
            * 'fit' : Array of the polynomial function coefficients of the
              pixel aberrations fit.
            * 'ph_fit' : Array of the polynomial function coefficients of
              the phase aberrations fit.
            * 'rel_err' : Vector of relative errors of the fit coefficients.
            * 'r_sq' : ``R**2`` goodness of fit.

        See Also:
            :func:`pyrost.AberrationsFit.fit` : Full details of the aberrations
            fitting algorithm.
        """
        if not self._isphase:
            raise ValueError("'phase' is not defined inside the container.")
        return self.get_fit(center=center, axis=axis).fit(max_order=max_order,
                                                            xtol=xtol, ftol=ftol,
                                                            loss=loss)

    def defocus_sweep(self, defoci_x: np.ndarray, defoci_y: Optional[np.ndarray]=None, size: int=51,
                      extra_args: Dict[str, Union[float, bool, str]]={}, return_extra: bool=False,
                      verbose: bool=True) -> Tuple[List[float], Dict[str, np.ndarray]]:
        r"""Calculate a set of reference images for each defocus in `defoci` and
        return an average R-characteristic of an image (the higher the value the
        sharper reference image is). Return the intermediate results if `return_extra`
        is True.

        Args:
            defoci_x : Array of defocus distances along the horizontal detector axis [m].
            defoci_y : Array of defocus distances along the vertical detector axis [m].
            size : Local variance filter size in pixels.
            extra_args : Extra arguments parser to the :func:`STData.get_st` and
                :func:`SpeckleTracking.update_reference` methods. The following
                keyword values are allowed:

                * 'ds_y' : Reference image sampling interval in pixels along the
                  horizontal axis. The default value is 1.0.
                * 'ds_x' : Reference image sampling interval in pixels along the
                  vertical axis. The default value is 1.0.
                * 'aberrations' : Add `pixel_aberrations` to `pixel_map` of
                  :class:`SpeckleTracking` object if it's True. The default value
                  is False.
                * 'ff_correction' : Apply dynamic flatfield correction if it's True.
                  The default value is False.
                * 'hval' : Kernel bandwidth in pixels for the reference image update.
                  The default value is 1.0.
                * 'ref_method' : Choose the reference image update algorithm. The
                  following keyword values are allowed:

                  * 'KerReg' : Kernel regression algorithm.
                  * 'LOWESS' : Local weighted linear regression.

                  The default value is 'KerReg'.

            return_extra : Return a dictionary with the intermediate results if True.

        Returns:
            A tuple of two items ('r_vals', 'extra'). The elements are as
            follows:

            * 'r_vals' : Array of the average values of `reference_image` gradients
              squared.
            * 'extra' : Dictionary with the intermediate results. Only if `return_extra`
              is True. Contains the following data:

              * reference_image : The generated set of reference profiles.
              * r_images : The set of local variance images of reference profiles.

        Notes:
            R-characteristic is called a local variance and is given by:

            .. math::
                R[i, j] = \frac{\sum_{i^{\prime} = -N / 2}^{N / 2}
                \sum_{j^{\prime} = -N / 2}^{N / 2} (I[i - i^{\prime}, j - j^{\prime}]
                - \bar{I}[i, j])^2}{\bar{I}^2[i, j]}

            where :math:`\bar{I}[i, j]` is a local mean and defined as follows:

            .. math::
                \bar{I}[i, j] = \frac{1}{N^2} \sum_{i^{\prime} = -N / 2}^{N / 2}
                \sum_{j^{\prime} = -N / 2}^{N / 2} I[i - i^{\prime}, j - j^{\prime}]

        See Also:
            :func:`pyrost.SpeckleTracking.update_reference` : reference image update
            algorithm.
        """
        if defoci_y is None:
            defoci_y = defoci_x.copy()

        ds_y = extra_args.get('ds_y', 1.0)
        ds_x = extra_args.get('ds_x', 1.0)
        aberrations = extra_args.get('aberrations', False)
        ff_correction = extra_args.get('ff_correction', False)
        hval = extra_args.get('hval', 1.0)
        ref_method = extra_args.get('ref_method', 'KerReg')

        r_vals = []
        extra = {'reference_image': [], 'r_image': []}
        kernel = np.ones(int(size)) / size
        st_obj = self.update_defocus(defoci_x.ravel()[0],
                                     defoci_y.ravel()[0]).get_st(ds_y=ds_y, ds_x=ds_x,
                                                                 aberrations=aberrations,
                                                                 ff_correction=ff_correction)

        for defocus_x, defocus_y in tqdm(zip(defoci_x.ravel(), defoci_y.ravel()),
                                           total=len(defoci_x), disable=not verbose,
                                           desc='Generating defocus sweep'):
            st_obj.di_pix *= np.abs(defoci_y.ravel()[0] / defocus_y)
            st_obj.dj_pix *= np.abs(defoci_x.ravel()[0] / defocus_x)
            st_obj.update_reference.inplace_update(hval=hval, method=ref_method)
            extra['reference_image'].append(st_obj.reference_image)
            mean = st_obj.reference_image.copy()
            mean_sq = st_obj.reference_image**2
            if st_obj.reference_image.shape[0] > size:
                mean = fft_convolve(mean, kernel, mode='reflect', axis=0,
                                    num_threads=self.num_threads)[size // 2:-size // 2]
                mean_sq = fft_convolve(mean_sq, kernel, mode='reflect', axis=0,
                                       num_threads=self.num_threads)[size // 2:-size // 2]
            if st_obj.reference_image.shape[1] > size:
                mean = fft_convolve(mean, kernel, mode='reflect', axis=1,
                                    num_threads=self.num_threads)[:, size // 2:-size // 2]
                mean_sq = fft_convolve(mean_sq, kernel, mode='reflect', axis=1,
                                       num_threads=self.num_threads)[:, size // 2:-size // 2]
            r_image = (mean_sq - mean**2) / mean**2
            extra['r_image'].append(r_image)
            r_vals.append(np.mean(r_image))

        if return_extra:
            return r_vals, extra
        return r_vals

    def get(self, attr: str, value: Optional[Any]=None) -> Any:
        """Return a dataset with `mask` and `roi` applied.
        Return `value` if the attribute is not found.

        Args:
            attr : Attribute to fetch.
            value : Return if `attr` is not found.

        Returns:
            `attr` dataset with `mask` and `roi` applied.
            `value` if `attr` is not found.
        """
        if attr in self:
            val = super(STData, self).get(attr)
            if val is not None:
                if self.protocol.get_is_data(attr):
                    val = val[..., self.roi[0]:self.roi[1], self.roi[2]:self.roi[3]]
                if attr in ['basis_vectors', 'flatfields', 'data', 'mask',
                            'pixel_translations', 'translations']:
                    val = val[self.good_frames]
            return val
        return value

    def get_st(self, ds_y: float=1.0, ds_x: float=1.0, aberrations: bool=False,
               ff_correction: bool=False) -> SpeckleTracking:
        """Return :class:`SpeckleTracking` object derived from the container.
        Return None if `defocus_x` or `defocus_y` doesn't exist in the container.

        Args:
            ds_y : Reference image sampling interval in pixels along the vertical
                axis.
            ds_x : Reference image sampling interval in pixels along the
                horizontal axis.
            aberrations : Add `pixel_aberrations` to `pixel_map` of
                :class:`SpeckleTracking` object if it's True.
            ff_correction : Apply dynamic flatfield correction if it's True.

        Returns:
            An instance of :class:`SpeckleTracking` derived from the container.
            None if `defocus_x` or `defocus_y` are not defined.
        """
        if not self._isdefocus:
            raise ValueError("'defocus_x' is not defined inside the container.")

        data = np.ascontiguousarray(self.get('mask') * self.get('data'))
        pixel_map = np.ascontiguousarray(self.get('pixel_map'))
        if aberrations:
            pixel_map += self.get('pixel_aberrations')
        whitefield = np.ascontiguousarray(self.get('whitefield'))
        if ff_correction:
            flatfields = self.get('flatfields')
            if not flatfields is None:
                np.rint(data * np.where(flatfields > 0, whitefield / flatfields, 1.),
                        out=data, casting='unsafe')
        dij_pix = np.ascontiguousarray(np.swapaxes(self.get('pixel_translations'), 0, 1))
        return SpeckleTracking(parent=ref(self), data=data, dj_pix=dij_pix[1], di_pix=dij_pix[0],
                               num_threads=self.num_threads, pixel_map=pixel_map, sigma=self.sigma,
                               ds_y=ds_y, ds_x=ds_x, whitefield=whitefield)

    def get_fit(self, center: int=0, axis: int=1) -> AberrationsFit:
        """Return an :class:`AberrationsFit` object for parametric regression
        of the lens' aberrations profile. Raises an error if 'defocus_x' or
        'defocus_y' is not defined.

        Args:
            center : Index of the zerro scattering angle or direct beam pixel.
            axis : Detector axis along which the fitting is performed.

        Raises:
            ValueError : If 'defocus_x' or 'defocus_y' is not defined in the
                container.

        Returns:
            An instance of :class:`AberrationsFit` class.
        """
        if not self._isphase:
            raise ValueError("'phase' or 'pixel_aberrations' are not defined inside the container.")

        data_dict = {attr: self.get(attr) for attr in AberrationsFit.attr_set if attr in self}
        if axis == 0:
            data_dict.update({attr: self.get(data_attr)
                              for attr, data_attr in AberrationsFit.y_lookup.items()})
        elif axis == 1:
            data_dict.update({attr: self.get(data_attr)
                              for attr, data_attr in AberrationsFit.x_lookup.items()})
        else:
            raise ValueError('invalid axis value: {:d}'.format(axis))

        data_dict['defocus'] = np.abs(data_dict['defocus'])
        if center <= self.roi[2 * axis]:
            data_dict['pixels'] = np.arange(self.roi[2 * axis],
                                            self.roi[2 * axis + 1]) - center
            data_dict['pixel_aberrations'] = data_dict['pixel_aberrations'][axis].mean(axis=1 - axis)
        elif center >= self.roi[2 * axis - 1] - 1:
            data_dict['pixels'] = center - np.arange(self.roi[2 * axis],
                                                     self.roi[2 * axis + 1])
            idxs = np.argsort(data_dict['pixels'])
            data_dict['pixel_aberrations'] = -data_dict['pixel_aberrations'][axis].mean(axis=1 - axis)[idxs]
            data_dict['pixels'] = data_dict['pixels'][idxs]
        else:
            raise ValueError('Origin must be outside of the region of interest')

        return AberrationsFit(parent=ref(self), **data_dict)

    def get_pca(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Perform the Principal Component Analysis [PCA]_ of the measured data and
        return a set of eigen flatfields (EFF).

        Returns:
            A tuple of ('cor_data', 'effs', 'eig_vals'). The elements are
            as follows:

            * 'cor_data' : Background corrected stack of measured frames.
            * 'effs' : Set of eigen flat-fields.
            * 'eig_vals' : Corresponding eigen values for each of the eigen
              flat-fields.

        References:
            .. [PCA] Vincent Van Nieuwenhove, Jan De Beenhouwer, Francesco De Carlo,
                    Lucia Mancini, Federica Marone, and Jan Sijbers, "Dynamic intensity
                    normalization using eigen flat fields in X-ray imaging," Opt.
                    Express 23, 27975-27989 (2015).
        """
        good_data, good_mask, good_wf = self.get('data'), self.get('mask'), self.get('whitefield')
        cor_data = np.zeros(good_data.shape, dtype=good_wf.dtype)
        np.subtract(good_data, good_wf, where=good_mask, out=cor_data)
        mat_svd = np.tensordot(cor_data, cor_data, axes=((1, 2), (1, 2)))
        eig_vals, eig_vecs = np.linalg.eig(mat_svd)
        effs = np.tensordot(eig_vecs, cor_data, axes=((0,), (0,)))
        return cor_data, effs, eig_vals / eig_vals.sum()

    @dict_to_object
    def update_flatfields(self, method: str='median', size: int=11,
                          cor_data: Optional[np.ndarray]=None,
                          effs: Optional[np.ndarray]=None) -> STData:
        """Return a new :class:`STData` object with a new set of flatfields.
        The flatfields are generated by the dint of median filtering or Principal
        Component Analysis [PCA]_.

        Args:
            method : Method to generate the flatfields. The following keyword
                values are allowed:

                * 'median' : Median `data` along the first axis.
                * 'pca' : Generate a set of flatfields based on eigen flatfields
                  `effs`. `effs` can be obtained with :func:`STData.get_pca` method.

            size : Size of the filter window in pixels used for the 'median' generation
                method.
            cor_data : Background corrected stack of measured frames.
            effs : Set of Eigen flatfields used for the 'pca' generation method.

        Raises:
            ValueError : If the `method` keyword is invalid.
            AttributeError : If the `whitefield` is absent in the :class:`STData`
                container when using the 'pca' generation method.
            ValuerError : If `effs` were not provided when using the 'pca' generation
                method.

        Returns:
            New :class:`STData` object with the updated `flatfields`.

        References:
            .. [PCA] Vincent Van Nieuwenhove, Jan De Beenhouwer, Francesco De Carlo,
                     Lucia Mancini, Federica Marone, and Jan Sijbers, "Dynamic
                     intensity normalization using eigen flat fields in X-ray
                     imaging," Opt. Express 23, 27975-27989 (2015).

        See Also:
            :func:`pyrost.STData.get_pca` : Method to generate eigen flatfields.
        """
        good_wf = self.get('whitefield')
        if method == 'median':
            good_data = self.get('data')
            outliers = np.abs(good_data - good_wf) < 3 * np.sqrt(good_wf)
            good_flats = median_filter(good_data, size=(size, 1, 1), mask=outliers,
                                       num_threads=self.num_threads)
        elif method == 'pca':
            if cor_data is None:
                good_data, good_mask = self.get('data'), self.get('mask')
                cor_data = np.zeros(good_data.shape, dtype=good_wf.dtype)
                np.subtract(good_data, good_wf, where=good_mask, out=cor_data)
            if effs is None:
                raise ValueError('No eigen flat fields were provided')

            weights = np.tensordot(cor_data, effs, axes=((1, 2), (1, 2))) / \
                      np.sum(effs * effs, axis=(1, 2))
            good_flats = np.asarray(np.tensordot(weights, effs, axes=((1,), (0,))) + good_wf,
                                    dtype=good_data.dtype)
        else:
            raise ValueError('Invalid method argument')

        flatfields = np.zeros(self.data.shape, dtype=good_data.dtype)
        flatfields[self.good_frames, self.roi[0]:self.roi[1],
                    self.roi[2]:self.roi[3]] = good_flats
        return {'flatfields': flatfields}

    def write_cxi(self, cxi_file: h5py.File) -> None:
        """Write all the `attr` to a CXI file `cxi_file`.

        Args:
            cxi_file : :class:`h5py.File` object of the CXI file.
            overwrite : Overwrite the content of `cxi_file` file if it's True.

        Raises:
            ValueError : If `overwrite` is False and the data is already present
                in `cxi_file`.
        """
        for attr, data in self.items():
            if attr in self.protocol and data is not None:
                self.protocol.write_cxi(attr, data, cxi_file)



"""pyrost.multislice (ms_sim) is capable to propagate the wavefront
through a bulky sample by the dint of multislice beam propagation algorithm.
The back-end is based on FFTW library.

(c) Nikolay Ivanov, 2020.
"""
from .ms_parameters import Element, Material, MSParams
from .mslice import MLL, MSPropagator



"""Multislice beam propagation simulation. Generate a profile of
the beam that propagates through a bulky sample.
:class:`pyrost.multislice.MLL` generates a transmission profile of
a wedged Multilayer Laue lens (MLL). :class:`pyrost.multislice.MSPropagator`
calculates the beam propagation and spits out the wavefield profile of
the propagated beam together with the transmission profile of the sample
at each slice.

Examples:

    Initialize a MLL object with the :class:`pyrost.multislice.MSParams`
    parameters object `params` with :func:`pyrost.multislice.MLL.import_params`.
    Then you can initialize a multislice beam propagator
    :class:`pyrost.multislice.MSPropagator` with `params` and `mll` and perform
    the multislice beam propagation as follows:

    >>> import pyrost.multislice as ms_sim
    >>> params = ms_sim.MSParams.import_default()
    >>> mll = ms_sim.MLL.import_params(params)
    >>> ms_prgt = ms_sim.MSPropagator(params, mll)
    >>> ms_prgt.beam_propagate() # doctest: +SKIP

    All the results are saved into `ms_prgt.beam_profile` and
    `ms_prgt.smp_profile` attributes.
"""
from __future__ import annotations
from multiprocessing import cpu_count
from typing import Iterable, Optional, Tuple, Union
from tqdm.auto import tqdm
import numpy as np
from .ms_parameters import MSParams
from ..data_container import DataContainer, dict_to_object
from ..bin import mll_profile, FFTW, empty_aligned, rsc_wp

class MLL(DataContainer):
    """
    Multilayer Laue lens class.

    Parameters
    ----------
    layers : numpy.ndarray
        MLL's bilayers x coordinates [um].
    mat1_r, mat2_r : complex
        Fresnel transmission coefficients for each of the materials,
        that the MLL's bilayers are composed of. The coefficients are the
        ratio of a wavefield propagating through a single slice of the
        lens.
    sigma : float
        Bilayer's interdiffusion length [um].

    Attributes
    ----------
    attr_set : set
        Set of the attributes in the container which are necessary
        to initialize in the constructor.
    """
    attr_set = {'layers', 'mat1_r', 'mat2_r', 'sigma'}
    en_to_wl = 1.239841929761768 # h * c / e [eV * um]

    def __init__(self, mat1_r: complex, mat2_r: complex, sigma: float,
                 layers: np.ndarray) -> None:
        super(MLL, self).__init__(mat1_r=mat1_r, mat2_r=mat2_r, sigma=sigma,
                                  layers=layers)

    @classmethod
    def import_params(cls, params: MSParams) -> MLL:
        """Return a new :class:`MLL` object import from the
        :class:`MSParams` multislice parameters object.

        Parameters
        ----------
        params : MSParams
            Experimental parameters of the multislice beam
            propagation simulation.

        Returns
        -------
        mll : MLL
            New :class:`MLL` object.
        """
        n_arr = np.arange(params.n_min, params.n_max)
        z_coords = params.z_step * np.arange(params.mll_depth // params.z_step)
        layers = np.sqrt(n_arr * params.focus * params.mll_wl + \
                              n_arr**2 * params.mll_wl**2 / 4)
        layers = layers * (1 - z_coords[:, None] / (2 * params.focus))
        mat1_r = params.get_mat1_r(cls.en_to_wl / params.wl)
        mat2_r = params.get_mat2_r(cls.en_to_wl / params.wl)
        sigma = params.mll_sigma
        return cls(mat1_r=mat1_r, mat2_r=mat2_r, sigma=sigma,
                   layers=layers)

    @property
    def n_slices(self) -> int:
        "Total number of slices"
        return self.layers.shape[0]

    @dict_to_object
    def update_interdiffusion(self, sigma: float) -> MLL:
        """Return a new :class:`MLL` object with the updated `sigma`.

        Parameters
        ----------
        sigma : float
            Bilayer's interdiffusion length [um].

        Returns
        -------
        mll : MLL
            New :class:`MLL` object with the updated `sigma`.
        """
        return {'sigma': sigma}

    @dict_to_object
    def update_materials(self, mat1_r: complex, mat2_r: complex) -> MLL:
        """Return a new :class:`MLL` object with the updated materials
        `mat1_r` and `mat2_r`.

        Parameters
        ----------
        mat1_r, mat2_r : complex
            Fresnel transmission coefficients for each of the materials,
            that the MLL's bilayers are composed of. The coefficients are the
            ratio of a wavefield propagating through a single slice of the
            lens.

        Returns
        -------
        mll : MLL
            New :class:`MLL` object with the updated `mat1_r` and `mat2_r`.
        """
        return {'mat1_r': mat1_r, 'mat2_r': mat2_r}

    def get_span(self) -> Tuple[float, float]:
        """Return the pair of bounds (x_min, x_max) of the MLL
        along the x axis.

        Returns
        -------
        x_min, x_max : float
            MLL's bounds along the x axis [um].
        """
        return (self.layers.min(), self.layers.max())

    def get_profile(self, x_arr: np.ndarray, output: np.ndarray,
                    num_threads: int=1) -> Iterable:
        """Return a generator, that yields transmission profiles of
        the lens at each slice and writes to `output` array.

        Parameters
        ----------
        x_arr : numpy.ndarray
            Coordinates array [um].
        output: numpy.ndarray
            Output array.
        num_threads : int, optional
            Number of threads.

        Returns
        -------
        slices : iterable
            The generator, which yields lens' transmission profiles.
        """
        for idx, layer in enumerate(self.layers):
            output[idx, :] = mll_profile(x_arr=x_arr, layers=layer, mt0=self.mat1_r,
                                         mt1=self.mat2_r, sigma=self.sigma,
                                         num_threads=num_threads)
            yield output[idx]

class MSPropagator(DataContainer):
    """One-dimensional Multislice beam propagation class.
    Generates beam profile, that propagates through the sample
    using multislice approach.

    Parameters
    ----------
    params : MSParams
        Experimental parameters
    sample : MLL
        Sample class, that generates the sample's
        transmission profile.
    num_threads : int, optional
        Number of threads used in the calculations.
    **kwargs : dict
        Attributes specified in `init_set`.

    Attributes
    ----------
    attr_set : set
        Set of attributes in the container which are necessary
        to initialize in the constructor.
    init_set : set
        Set of optional data attributes.

    Notes
    -----
    Necessary attributes:

    * sample : Sample's transmission profile generator.
    * num_threads : Number of threads used in the calculations.
    * params : Experimental parameters.

    Optional attributes:

    * beam_profile : Beam profiles at each slice.
    * fx_arr : Spatial frequencies array [um^-1].
    * kernel : Diffraction kernel.
    * smp_profile : Sample's transmission profile at each slice.
    * x_arr : Coordinates array in the transverse plane [um].
    * wf_inc : Wavefront at the entry surface.
    * z_arr : Coordinates array along the propagation axis [um].
    """
    attr_set = {'params', 'sample'}
    init_set = {'beam_profile', 'fx_arr', 'kernel', 'num_threads', 'smp_profile', 'x_arr',
                'wf_inc', 'z_arr'}

    inits = {'num_threads'  : lambda obj: np.clip(1, 64, cpu_count()),
             'x_arr'        : lambda obj: obj.params.get_xcoords(),
             'z_arr'        : lambda obj: obj.params.get_zcoords(),
             'fx_arr'       : lambda obj: np.fft.fftfreq(obj.size, obj.params.x_step),
             'kernel'       : lambda obj: obj.params.get_kernel(obj.fx_arr) / obj.fx_arr.size,
             'wf_inc'       : lambda obj: obj._inc_wavefront()}

    def __init__(self, params: MSParams, sample: MLL, num_threads: Optional[int]=None,
                 **kwargs: np.ndarray) -> None:
        super(MSPropagator, self).__init__(params=params, sample=sample,
                                           num_threads=num_threads, **kwargs)

    def _inc_wavefront(self):
        wf_inc = np.ones(self.x_arr.shape, dtype=np.complex128)
        x_min, x_max = self.sample.get_span()
        wf_inc[(self.x_arr < x_min) | (self.x_arr > x_max)] = 0.0
        return wf_inc

    @property
    def size(self) -> int:
        "Number of points in a single slice."
        return self.x_arr.size

    @dict_to_object
    def update_inc_wavefront(self, wf_inc: np.ndarray) -> MSPropagator:
        """Return a new :class:`MSPropagator` object with the updated `wf_inc`.

        Parameters
        ----------
        wf_inc : numpy.ndarray
            Wavefront at the entry surface.

        Returns
        -------
        MSPropagator
            A new :class:`MSPropagator` object with the updated `wf_inc`.
        """
        if wf_inc.shape != self.x_arr.shape:
            raise ValueError(f'Wavefront shape must be equal to {self.x_arr.shape:s}')
        if wf_inc.dtype != np.complex128:
            raise ValueError("Wavefront datatype must be 'complex128'")
        return {'beam_profile': None, 'wf_inc': wf_inc}

    def generate_sample(self, verbose: bool=True) -> None:
        """Generate the transmission profile of the sample. The results are
        written to `smp_profile` attribute.

        Parameters
        ----------
        verbose : bool, optional
            Set verbosity of the computation process.
        """
        self.smp_profile = empty_aligned((self.sample.n_slices, self.x_arr.size),
                                         dtype='complex128')
        itor = self.sample.get_profile(x_arr=self.x_arr, output=self.smp_profile,
                                       num_threads=self.num_threads)

        if verbose:
            itor = tqdm(enumerate(itor), total=self.sample.n_slices,
                        bar_format='{desc} {percentage:3.0f}% {bar} Slice {n_fmt} / {total_fmt} '\
                                '[{elapsed}<{remaining}, {rate_fmt}{postfix}]')
        for idx, _ in itor:
            if verbose:
                itor.set_description(f"z = {self.z_arr[idx]:.2f} um")

    def beam_propagate(self, verbose: bool=True) -> None:
        """Perform the multislice beam propagation. The results are
        written to `beam_profile` attribute. Generates `smp_profile`
        attribute if it wasn't initialized before.

        Parameters
        ----------
        verbose : bool, optional
            Set verbosity of the computation process.
        """
        self.beam_profile = empty_aligned((self.sample.n_slices + 1, self.x_arr.size),
                                          dtype='complex128')
        current_slice = empty_aligned(self.x_arr.size, dtype='complex128')
        self.beam_profile[0] = self.wf_inc
        current_slice[:] = self.wf_inc

        fft_obj = FFTW(current_slice, current_slice, flags=('FFTW_ESTIMATE',),
                       threads=self.num_threads)
        ifft_obj = FFTW(current_slice, self.beam_profile[1], direction='FFTW_BACKWARD',
                        flags=('FFTW_ESTIMATE',), threads=self.num_threads)

        if self.smp_profile is None:
            self.smp_profile = empty_aligned((self.sample.n_slices, self.x_arr.size),
                                             dtype='complex128')
            itor = self.sample.get_profile(x_arr=self.x_arr, output=self.smp_profile,
                                           num_threads=self.num_threads)
        else:
            itor = (layer for layer in self.smp_profile)

        miniters = max(self.sample.n_slices // 100, 1)
        itor = tqdm(itor, total=self.sample.n_slices, miniters=miniters, disable=not verbose,
                    bar_format='{desc} {percentage:3.0f}% {bar} Slice {n_fmt} / {total_fmt} '\
                               '[{elapsed}<{remaining}, {rate_fmt}{postfix}]')

        for idx, layer in enumerate(itor):
            if verbose and not idx % miniters:
                itor.set_description(f"z = {self.z_arr[idx]:.2f} um")
            current_slice *= layer
            fft_obj.execute()
            current_slice *= self.kernel
            ifft_obj.update_arrays(current_slice, self.beam_profile[idx + 1])
            ifft_obj.execute()
            current_slice[:] = ifft_obj.output_array

    def beam_downstream(self, z_arr: np.ndarray, step: Optional[float]=None,
                        return_coords: bool=True, verbose: bool=True,
                        backend: str='fftw') -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """Return the wavefront at distance `z_arr` downstream from
        the exit surface using Rayleigh-Sommerfeld convolution.

        Parameters
        ----------
        z_arr : numpy.ndarray
            Array of distances [um].
        step : float, optional
            Sampling interval of the downstream coordinate array [um].
            Equals to the sampling interval at the exit surface if it's
            None.
        return_coords : bool, optional
            Return the coordinates array of the downstream plane.
        verbose : bool, optional
            Set verbosity of the computation process.

        Returns
        -------
        wavefronts : numpy.ndarray
            Set of wavefronts calculated at the distance `z_arr`
            downstream from the exit surface.
        x_arr : numpy.ndarray
            Array of coordinates at the plane downstream [um].
            Only if `return_coords` is True.

        Raises
        ------
        AttributeError
            If `beam_profile` has not been generated. Call
            :func:`MSPropagator.beam_propagate` to generate.
        """
        if self.beam_profile is None:
            raise AttributeError('The beam profile has not been generated')
        if step is None:
            step = self.params.x_step
        size = int(self.x_arr.size**2 * max(self.params.x_step, step) * \
                   (step + self.params.x_step) / self.params.wl / np.min(z_arr))
        wf0 = self.beam_profile[-1]
        if size > self.x_arr.size:
            wf0 = np.pad(wf0, ((size - self.x_arr.size) // 2, (size - self.x_arr.size) // 2))
        else:
            size = self.x_arr.size

        itor = np.atleast_1d(z_arr)
        if verbose:
            itor = tqdm(itor, total=itor.size)

        wavefronts = []
        for dist in itor:
            if verbose:
                itor.set_description(f"z = {dist:.2f} um")
            wavefronts.append(rsc_wp(wft=wf0, dx0=self.params.x_step, dx=step,
                                     z=dist, wl=self.params.wl, backend=backend,
                                     num_threads=self.num_threads))
        if return_coords:
            x_arr = step * np.arange(-size // 2, size - size // 2) + np.mean(self.x_arr)
            return np.squeeze(np.stack(wavefronts, axis=1)), x_arr
        return np.squeeze(np.stack(wavefronts, axis=1))



cimport numpy as np

ctypedef int (*rconvolve_func)(double*, double*, int, unsigned long*, double*,
                               unsigned long, int, int, double, unsigned)
ctypedef int (*cconvolve_func)(double complex*, double complex*, int, unsigned long*, double complex*,
                               unsigned long, int, int, double complex, unsigned)

cdef extern from "pocket_fft.h":
    unsigned long next_fast_len_fftw(unsigned long target) nogil
    unsigned long good_size(unsigned long n) nogil

cdef extern from "fft_functions.h":
    int rfft_convolve_fftw(double *out, double *inp, int ndim, unsigned long* dims, double *krn,
                           unsigned long ksize, int axis, int mode, double cval, unsigned threads) nogil

    int cfft_convolve_fftw(double complex *out, double complex *inp, int ndim, unsigned long* dims,
                           double complex *krn, unsigned long ksize, int axis, int mode, double complex cval,
                           unsigned threads) nogil

    int rfft_convolve_np(double *out, double *inp, int ndim, unsigned long* dims, double *krn,
                          unsigned long ksize, int axis, int mode, double cval, unsigned threads) nogil

    int cfft_convolve_np(double complex *out, double complex *inp, int ndim, unsigned long* dims,
                         double complex *krn, unsigned long ksize, int axis, int mode, double complex cval,
                         unsigned threads) nogil

    int rsc_np(double complex *out, double complex *inp, int ndim, unsigned long *dims,
               int axis, double dx0, double dx, double z, double wl, unsigned threads) nogil

    int rsc_fftw(double complex *out, double complex *inp, int ndim, unsigned long *dims,
                 int axis, double dx0, double dx, double z, double wl, unsigned threads) nogil

    int fraunhofer_np(double complex *out, double complex *inp, int ndim, unsigned long *dims, int axis,
                      double dx0, double dx, double z, double wl, unsigned threads) nogil

    int fraunhofer_fftw(double complex *out, double complex *inp, int ndim, unsigned long *dims, int axis,
                        double dx0, double dx, double z, double wl, unsigned threads) nogil

    int gauss_kernel1d(double *out, double sigma, unsigned order, unsigned long ksize, int step) nogil

    int gauss_filter_r(double *out, double *inp, int ndim, unsigned long *dims, double *sigma,
                       unsigned *order, int mode, double cval, double truncate, unsigned threads,
                       rconvolve_func fft_convolve) nogil

    int gauss_filter_c(double complex *out, double complex *inp, int ndim, unsigned long *dims,
                       double *sigma, unsigned *order, int mode, double complex cval, double truncate,
                       unsigned threads, cconvolve_func fft_convolve) nogil

    int gauss_grad_mag_r(double *out, double *inp, int ndim, unsigned long *dims, double *sigma,
                         int mode, double cval, double truncate, unsigned threads,
                         rconvolve_func fft_convolve) nogil

    int gauss_grad_mag_c(double *out, double complex *inp, int ndim, unsigned long *dims,
                         double *sigma, int mode, double complex cval, double truncate, unsigned threads,
                         cconvolve_func fft_convolve) nogil

cdef extern from "routines.h":
    void barcode_bars(double *bars, unsigned long size, double x0, double b_dx, double rd, long seed) nogil

    int ml_profile(complex *out, double *inp, unsigned long isize, double *layers, unsigned long lsize, 
                   complex mt0, complex mt1, complex mt2, double sgm, unsigned threads) nogil

    int frames(double *out, double *pfx, double *pfy, double dx, double dy, unsigned long *ishape,
               unsigned long *oshape, long seed, unsigned threads) nogil

cdef extern from "median.h":
    int compare_double(void *a, void *b) nogil
    int compare_float(void *a, void *b) nogil
    int compare_int(void *a, void *b) nogil
    int compare_uint(void *a, void *b) nogil
    int compare_ulong(void *a, void *b) nogil

    int median_c "median" (void *out, void *data, unsigned char *mask, int ndim, unsigned long *dims,
                 unsigned long item_size, int axis, int (*compar)(void*, void*), unsigned threads) nogil

    int median_filter_c "median_filter" (void *out, void *data, unsigned char *mask, int ndim,
                        unsigned long *dims, unsigned long item_size, unsigned long *fsize, int mode,
                        void *cval, int (*compar)(void*, void*), unsigned threads) nogil

cdef extern from "fftw3.h":
    void fftw_init_threads() nogil
    void fftw_cleanup_threads() nogil

cdef enum:
    EXTEND_CONSTANT = 0
    EXTEND_NEAREST = 1
    EXTEND_MIRROR = 2
    EXTEND_REFLECT = 3
    EXTEND_WRAP = 4

cdef int extend_mode_to_code(str mode) except -1
cdef np.ndarray check_array(np.ndarray array, int type_num)
cdef np.ndarray number_to_array(object num, np.npy_intp rank, int type_num)
cdef np.ndarray normalize_sequence(object inp, np.npy_intp rank, int type_num)
cdef np.ndarray ml_profile_wrapper(np.ndarray x_arr, np.ndarray layers, complex mt0,
                                   complex mt1, complex mt2, double sigma, unsigned num_threads)
